import logging

from pyspedas.tplot_tools import time_datetime
from pyspedas.tplot_tools import store_data, options


def load(trange=None,
         collection=None,
         measurements=None,
         models=None,
         sampling_step=None,
         auxiliaries=None,
         residuals=False):
    """

    """

    try:
        from viresclient import SwarmRequest
    except ImportError:
        logging.info('The viresclient package is needed for this operation, but does not appear to be installed.')
        logging.info('To use this feature, install viresclient with "pip install viresclient".')
        logging.info('If pip install fails, try "conda install viresclient".')

    from .config import CONFIG

    if CONFIG['access_token'] == '':
        logging.error('Invalid access token')
        return
    else:
        access_token = CONFIG['access_token']

    if trange is None:
        logging.error('No time range specified')
        return

    tr = time_datetime(trange)

    if not isinstance(measurements, list):
        measurements = [measurements]

    if auxiliaries is not None and not isinstance(auxiliaries, list):
        auxiliaries = [auxiliaries]

    if models is not None:
        if not isinstance(models, list):
            models = [models]

    request = SwarmRequest(url="https://vires.services/ows", token=access_token)
    if isinstance(collection, list):
        request.set_collection(*collection)
    else:
        request.set_collection(collection)

    if auxiliaries is None:
        request.set_products(measurements=measurements, models=models, sampling_step=sampling_step, residuals=residuals)
    else:
        request.set_products(measurements=measurements, auxiliaries=auxiliaries, models=models, sampling_step=sampling_step, residuals=residuals)

    data = request.get_between(start_time=tr[0], end_time=tr[1])
    return xarray_to_tplot(data.as_xarray())


def xarray_to_tplot(xr):
    out = []
    for key in xr.keys():
        times = xr[key].coords['Timestamp'].to_numpy()
        saved = store_data(key, data={'x': times, 'y': xr[key].data})
        options(key, 'ytitle', xr[key].description)
        options(key, 'ysubtitle', '[' + xr[key].units + ']')

        # find the legend if this is a vector
        for item in xr[key].coords:
            if item != 'Timestamp':
                options(key, 'legend_names', xr[key].coords[item].values.tolist())

        if saved:
            out.append(key)
        else:
            logging.warning('Problem saving: ' + key)
    return out



import unittest
from unittest.mock import patch, MagicMock

class TestImportErrorHandling(unittest.TestCase):
    def setUp(self):

        # Mocking only the logging part of pyspedas to capture logging outputs
        self.logging_patch = patch('pyspedas.logging.error', new_callable=MagicMock)
        self.mock_logging_error = self.logging_patch.start()

    def tearDown(self):
        self.logging_patch.stop()

    def test_pyspedas_mth5_import_error(self):
        import sys

        # Patch 'mth5' to raise ImportError
        with patch.dict('sys.modules', {'mth5': None}):
            if 'pyspedas.mth5' in sys.modules:
                del sys.modules['pyspedas.mth5']

            # Attempt to import 'pyspedas.mth5' which should raise ImportError due to 'mth5' not being available
            with self.assertRaises((ImportError, ModuleNotFoundError)):
                import pyspedas.mth5

            # Verify the error log was called with the expected message
            self.mock_logging_error.assert_any_call('MTH5 must be installed to use module pyspedas.mth5.')
            self.mock_logging_error.assert_any_call("Please install it using: pip install mth5")


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas
import time

sleep_time = 5

class LoadTestCases(unittest.TestCase):
    def test_load_pws_data(self):
        out_vars = pyspedas.projects.akebono.pws(time_clip=True)
        self.assertTrue(data_exists('akb_pws_RX1'))
        self.assertTrue(data_exists('akb_pws_RX2'))
        time.sleep(sleep_time)


    def test_load_pws_data_prefix_none(self):
        out_vars = pyspedas.projects.akebono.pws(prefix=None)
        self.assertTrue(data_exists('akb_pws_RX1'))
        self.assertTrue(data_exists('akb_pws_RX2'))
        time.sleep(sleep_time)

    def test_load_pws_data_suffix_none(self):
        out_vars = pyspedas.projects.akebono.pws(suffix=None)
        self.assertTrue(data_exists('akb_pws_RX1'))
        self.assertTrue(data_exists('akb_pws_RX2'))
        time.sleep(sleep_time)

    def test_load_pws_data_prefix_suffix(self):
        out_vars = pyspedas.projects.akebono.pws(prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_akb_pws_RX1_suf'))
        self.assertTrue(data_exists('pre_akb_pws_RX2_suf'))
        time.sleep(sleep_time)

    def test_load_rdm_data(self):
        out_vars = pyspedas.projects.akebono.rdm()
        self.assertTrue(data_exists('akb_L'))
        self.assertTrue(data_exists('akb_MLT'))
        self.assertTrue(data_exists('akb_rdm_FEIO'))
        time.sleep(sleep_time)

    def test_load_rdm_data_prefix_none(self):
        out_vars = pyspedas.projects.akebono.rdm(prefix=None)
        self.assertTrue(data_exists('akb_L'))
        self.assertTrue(data_exists('akb_MLT'))
        self.assertTrue(data_exists('akb_rdm_FEIO'))
        time.sleep(sleep_time)

    def test_load_rdm_data_suffix_none(self):
        out_vars = pyspedas.projects.akebono.rdm(suffix=None)
        self.assertTrue(data_exists('akb_L'))
        self.assertTrue(data_exists('akb_MLT'))
        self.assertTrue(data_exists('akb_rdm_FEIO'))
        time.sleep(sleep_time)

    def test_load_rdm_data_prefix_suffix(self):
        out_vars = pyspedas.projects.akebono.rdm(prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_akb_L_suf'))
        self.assertTrue(data_exists('pre_akb_MLT_suf'))
        self.assertTrue(data_exists('pre_akb_rdm_FEIO_suf'))
        time.sleep(sleep_time)

    def test_load_orb_data(self):
        out_vars = pyspedas.projects.akebono.orb()
        self.assertTrue(data_exists('akb_orb_geo'))
        self.assertTrue(data_exists('akb_orb_gdlat'))
        self.assertTrue(data_exists('akb_orb_gdlon'))
        time.sleep(sleep_time)

    def test_load_orb_prefix_none(self):
        out_vars = pyspedas.projects.akebono.orb(prefix=None)
        self.assertTrue(data_exists('akb_orb_geo'))
        self.assertTrue(data_exists('akb_orb_gdlat'))
        self.assertTrue(data_exists('akb_orb_gdlon'))
        time.sleep(sleep_time)

    def test_load_orb_suffix_none(self):
        out_vars = pyspedas.projects.akebono.orb(suffix=None)
        self.assertTrue(data_exists('akb_orb_geo'))
        self.assertTrue(data_exists('akb_orb_gdlat'))
        self.assertTrue(data_exists('akb_orb_gdlon'))
        time.sleep(sleep_time)

    def test_load_orb_prefix_suffix(self):
        out_vars = pyspedas.projects.akebono.orb(prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_akb_orb_geo_suf'))
        self.assertTrue(data_exists('pre_akb_orb_gdlat_suf'))
        self.assertTrue(data_exists('pre_akb_orb_gdlon_suf'))
        time.sleep(sleep_time)

    def test_load_notplot(self):
        out_vars = pyspedas.projects.akebono.pws(notplot=True)
        self.assertTrue('akb_pws_RX1' in out_vars)
        time.sleep(sleep_time)

    def test_downloadonly(self):
        files = pyspedas.projects.akebono.pws(downloadonly=True, trange=['2012-10-01', '2012-10-02'])
        self.assertTrue(os.path.exists(files[0]))
        time.sleep(sleep_time)


if __name__ == '__main__':
    unittest.main()

    


from getpass import getpass
from scipy.io import readsav
import requests
import os
import pickle
import logging
import warnings


def mms_login_lasp(always_prompt=False, headers={}):
    """
    This function logs the user into the SDC and returns a tuple with: (requests.Session object, username)
    """
    homedir = os.path.expanduser("~")
    user_input_passwd = False
    saved_auth = None

    # try to read saved pickle
    try:
        auth_file = open(os.sep.join([homedir, "mms_auth_info.pkl"]), "rb")
        saved_auth = pickle.load(auth_file)
        auth_file.close()
    except FileNotFoundError:
        pass

    # try to read the IDL sav file
    try:
        idl_auth_info = readsav(os.sep.join([homedir, "mms_auth_info.sav"]))
        saved_auth = {
            "user": idl_auth_info["auth_info"][0][0].decode("utf-8"),
            "passwd": idl_auth_info["auth_info"][0][1].decode("utf-8"),
        }
    except:
        pass

    if saved_auth is None or always_prompt:
        try:
            user = input("SDC username (blank for public access): ")
        except:
            logging.error(
                "Error while reading SDC username/password; defaulting to public user..."
            )
            user = ""

        if user != "":
            passwd = getpass()
        else:
            passwd = ""

        user_input_passwd = True
    else:
        user = saved_auth["user"]
        passwd = saved_auth["passwd"]

    session = requests.Session()

    if user != "":
        session.auth = (user, passwd)

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=ResourceWarning)
                testget = session.get(
                    "https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/download/science",
                    verify=True,
                    timeout=5,
                    headers=headers,
                )
        except:
            return session, None

        # check if the login failed
        if testget.status_code == 401:
            logging.error('Invalid password for user: ' + str(user) + '; trying public access..')
            user = None

    # only save the user's user/passwd if login was successful
    if user_input_passwd and user is not None:
        saved_auth = pickle.dump(
            {"user": user, "passwd": passwd},
            open(os.sep.join([homedir, "mms_auth_info.pkl"]), "wb"),
        )

    if user == "":
        user = None

    return session, user



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import logging
import pyspedas
import numpy as np

def tres(tplot_var):
    r"""
    Returns the time resolution of a tplot variable, defined as the median value 
    of the differences between time values from the data points, e.g., median(d.x[1:`*`]-d.x).
    Can be used for multiple variables.

    Parameters
    ----------
    - tplot_var (int or str or list): A tplot variable name or number, or a list of such names.

    Returns
    -------
    - delta_t (float or list): The time resolution, a median value. If the data variable 
        does not exist, or does not return a median, then the result is -1. If multiple 
        tplot variables are provided, returns a list of median values.

    Examples
    --------
    >>> import pyspedas
    >>> import pyspedas
    >>> import pyspedas
    >>> fgm_vars = pyspedas.projects.themis.fgm(probe='d', trange=['2013-11-5', '2013-11-6'])
    >>> dt = pyspedas.tres('thd_fgs_gse')
    >>> dts = pyspedas.tres(fgm_vars)

    """
    if isinstance(tplot_var, int):
        tplot_var = list(pyspedas.tplot_tools.data_quants.keys())[tplot_var]

    if isinstance(tplot_var, str):
        data = pyspedas.get_data(tplot_var)
        if data is not None:
            delta_t = np.median(data.times[1:-1]-data.times[0:-2])
        else:
            delta_t = -1

        return delta_t
    
    elif isinstance(tplot_var, list):
        delta_t = []
        for name in tplot_var:
            delta_t.append(tres(name))

        return delta_t
    
    else:
        logging.info("tres input must be a int, str or list!")
        return


import numpy as np
import pyspedas
import logging


def lineplot(var_data,
             var_times,
             this_axis,
             line_opts,
             yaxis_options,
             plot_extras,
             running_trace_count=None,
             time_idxs=None,
             style=None,
             var_metadata=None):
    """
    Generate a matplotlib line plot from a tplot variable

    Parameters
    ----------
        var_data: dict
            The data to be plotted (may have multiple traces)
        var_times:
            Array of datetime objects to use for x axis
        this_axis:
            The current axis (plot panel) we're working with
        line_opts: dict
            A dictionary of line options
        yaxis_options: dict
            A dictionary of y axis options
        plot_extras: dict
            A dictionary of 'extra' options (colors, etc)
        running_trace_count:
            If not Null, an integer representing the number of traces already processed in this pseudovariable. Defaults to None.
        time_idxs: np.ndarray
            If provided, an integer array specifying the subset of time indices to be plotted. Defaults to None.
        style:
            A matplotlib style to be used in the plot. Defaults to None.
        var_metadata: dict
            The metadata dictionary associated with this tplot variable (used as a fallback for trace labels). Defaults to None.

    Returns
    -------
        True

    """
    alpha = plot_extras.get('alpha')

    if len(var_data.y.shape) == 1:
        num_lines = 1
    else:
        num_lines = var_data.y.shape[1]

    is_errorbar_plot = False
    if 'dy' in var_data._fields:
        is_errorbar_plot = True

    if yaxis_options.get('legend_names') is not None:
        labels = yaxis_options['legend_names']
        labels = get_trace_options(labels, running_trace_count, num_lines)

        if labels[0] is None:
            labels = None
    else:
        labels = None
        if var_metadata.get('CDF') is not None:
            labels = var_metadata['CDF'].get('LABELS')

    legend_location = yaxis_options.get('legend_location')

    bbox_to_anchor = None
    if legend_location is not None:
        if legend_location == 'spedas':
            # the spedas legend puts the legend on the outside of the panel
            # to the right of the panel (just like in IDL)
            legend_location = 'center left'
            bbox_to_anchor = (1.04, 0.5)
    else:
        legend_location = 'upper right'

    legend_size = yaxis_options.get('legend_size')
    legend_shadow = yaxis_options.get('legend_shadow')
    legend_title = yaxis_options.get('legend_title')
    legend_titlesize = yaxis_options.get('legend_titlesize')
    legend_color = yaxis_options.get('legend_color')
    legend_markerfirst = yaxis_options.get('legend_markerfirst')
    legend_markerscale = yaxis_options.get('legend_markerscale')
    legend_linewidth = yaxis_options.get('legend_linewidth')
    legend_edgecolor = yaxis_options.get('legend_edgecolor')
    legend_facecolor = yaxis_options.get('legend_facecolor')
    legend_frameon = yaxis_options.get('legend_frameon')
    legend_ncols = yaxis_options.get('legend_ncols')
    if legend_ncols is None:
        legend_ncols = 1

    if legend_linewidth is None:
        legend_linewidth = 4

    if legend_size is None:
        legend_size = pyspedas.tplot_tools.tplot_opt_glob.get('charsize')

    markers = None
    if line_opts.get('marker') is not None:
        markers = line_opts['marker']
        markers = get_trace_options(markers, running_trace_count, num_lines, repeat=True)

    colors = None
    if plot_extras.get('line_color') is not None:
        colors = plot_extras['line_color']
    else:
        if style is None:
            if num_lines == 1:
                colors = ['k']
            elif num_lines == 2:
                colors = ['r', 'g']
            elif num_lines == 3:
                colors = ['b', 'g', 'r']
            elif num_lines == 4:
                colors = ['b', 'g', 'r', 'k']
            else:
                colors = ['k', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']
    colors = get_trace_options(colors, running_trace_count, num_lines, repeat=True)

    # line thickness
    if line_opts.get('line_width') is not None:
        thick = line_opts['line_width']
    else:
        thick = [0.5]
    thick = get_trace_options(thick, running_trace_count, num_lines, repeat=True)

    # line style
    if line_opts.get('line_style_name') is not None:
        line_style_user = line_opts['line_style_name']

        # line_style_user should already be a list
        # handle legacy values
        line_style = []
        for linestyle in line_style_user:
            if linestyle == 'solid_line':
                line_style.append('solid')
            elif linestyle == 'dot':
                line_style.append('dotted')
            elif linestyle == 'dash':
                line_style.append('dashed')
            elif linestyle == 'dash_dot':
                line_style.append('dashdot')
            else:
                line_style.append(linestyle)
    else:
        line_style = ['solid']
    line_style = get_trace_options(line_style, running_trace_count, num_lines, repeat=True)

    symbols = False
    if line_opts.get('symbols') is not None:
        if line_opts['symbols']:
            symbols = True

    # create the plot
    line_options = {'alpha': alpha}

    marker_every = None
    if line_opts.get('markevery') is not None:
        marker_every = line_opts['markevery']
        marker_every = get_trace_options(marker_every, running_trace_count, num_lines, repeat=True)

    marker_sizes = None
    if line_opts.get('marker_size') is not None:
        marker_sizes = line_opts['marker_size']
        marker_sizes = get_trace_options(marker_sizes, running_trace_count, num_lines, repeat=True)

    # check for error data first
    if is_errorbar_plot:
        # error data provided
        line_options['yerr'] = var_data.dy[time_idxs]
        plotter = this_axis.errorbar
        if line_opts.get('ecolor') is not None:
            line_options['ecolor'] = line_opts['ecolor']
        if line_opts.get('elinewidth') is not None:
            line_options['elinewidth'] = line_opts['elinewidth']
        if line_opts.get('errorevery') is not None:
            line_options['errorevery'] = line_opts['errorevery']
        if line_opts.get('capsize') is not None:
            line_options['capsize'] = line_opts['capsize']
    else:
        # no error data provided
        plotter = this_axis.plot
        # Note: to turn off connecting lines in an error bar plot, do not use the
        # 'symbols' option.  Instead, set the line_options metadata to 'None' (as a string).
        if symbols:
            plotter = this_axis.scatter

    for line in range(0, num_lines):
        if colors is not None:
            color = colors[line]
        else:
            color = None

        if markers is not None:
            marker = markers[line]
        else:
            marker = None

        if marker_sizes is not None:
            # Note: scaling of marker sizes in scatter plots and line plots is different!
            # For line plot and scatter plot marker sizes to match, the line plot
            # marker size should be the square root of the scatter plot marker size.
            # Maybe that should be enforced here....???

            if symbols:
                line_options['s'] = marker_sizes[line]
            else:
                line_options['markersize'] = marker_sizes[line]

        if symbols:
            this_line_style='None'
        else:
            this_line_style=line_style[line]

        if marker_every is not None:
            line_options['markevery'] = marker_every[line]

        this_line = plotter(var_times, var_data.y[time_idxs] if num_lines == 1 else var_data.y[time_idxs, line], color=color,
                            linestyle=this_line_style, linewidth=thick[line], marker=marker, **line_options)

        if labels is not None:
            try:
                if isinstance(this_line, list):
                    this_line[0].set_label(labels[line])
                else:
                    this_line.set_label(labels[line])
            except IndexError:
                continue

    if labels is not None:
        legend = this_axis.legend(loc=legend_location, fontsize=legend_size, shadow=legend_shadow, title=legend_title,
                         labelcolor=legend_color, markerfirst=legend_markerfirst, markerscale=legend_markerscale,
                         facecolor=legend_facecolor, edgecolor=legend_edgecolor, frameon=legend_frameon, ncols=legend_ncols,
                         title_fontsize=legend_titlesize, bbox_to_anchor=bbox_to_anchor)
        try:
            handles = legend.legend_handles
        except AttributeError:
            handles = legend.legendHandles
        for legobj in handles:
            legobj.set_linewidth(legend_linewidth)

    return True

def get_trace_options(parent_array, start_trace=None, num_traces=1, repeat=False, fill=False, fillval=None):
    """ Get options for a set of traces from a parent array, extending or slicing as necessary to handle pseudovariable options

    Parameters
    -----------
        parent_array: str or list of str
            An array of option values to select from
        start_trace: int
            If set, we are processing a pseuodovariable, and this is the count of line traces processed so far for
            previous sub-variables in the current pseuodovariable.

            If parent_array is long enough (e.g. if set on the pseudovariable and intended to cover the complete
            set of traces), we take a slice from start_trace with num_traces entries.  If it matches
            the number of traces requested (e.g. parent array comes from this subvariable), we take it as-is.
            If there are fewer entries than num_traces (e.g. a single value intended to apply to all traces),
            we repeat parent_array or add fill until there are enough values to take a slice from start_trace.

            If None, we're just processing a regular variable, so we take values starting at zero (extending or filling
            as necessary)
            Default: None
        num_traces: int
            The number of traces in the current (sub-)variable.
        repeat: bool
            If True, extend the parent array by repetition if necessary. Defaults to False.
        fill: bool
            If True, extend the parent array by adding fill values. Defaults to False.
        fillval: Any
            If fill=True, values to append to parent_array to make enough entries. Defaults to None.

    Returns:
    --------
        list of option values with num_traces entries

    """
    if not isinstance(parent_array,list):
        parent_array=[parent_array]
    parent_length = len(parent_array)
    output_array = parent_array
    if start_trace is not None:
        end_trace = start_trace + num_traces
        if parent_length >= start_trace+num_traces:
            output_array = parent_array[start_trace:end_trace]
        elif parent_length == num_traces:
            output_array = parent_array
        else:
            if repeat:
                expansion_factor = int((end_trace/parent_length + 1))
                expanded_array = np.tile(parent_array,expansion_factor)
                output_array = expanded_array[start_trace:end_trace]
            elif fill:
                output_array = parent_array
                missing=num_traces-parent_length
                output_array.extend(np.tile([fillval],missing))
            else:
                logging.warning("Length of trace options (%d) smaller than number of traces (%d)",parent_length, num_traces)
    else:
        if len(parent_array) >= num_traces:
            output_array = parent_array[0:num_traces]
        else:
            if repeat:
                expansion_factor = int(num_traces/parent_length + 1)
                expanded_array = np.tile(parent_array,expansion_factor)
                output_array = expanded_array[0:num_traces]
            elif fill:
                output_array = parent_array
                missing = num_traces - parent_length
                output_array.extend(np.tile([fillval], missing))
            else:
                logging.warning("Length of trace options (%d) smaller than number of traces (%d)",parent_length, num_traces)
    return output_array


import os
import pickle

MMS_AUTH_U = os.getenv('MMS_AUTH_U')
MMS_AUTH_P = os.getenv('MMS_AUTH_P')
pickle.dump({'user': MMS_AUTH_U, 'passwd': MMS_AUTH_P}, open(os.sep.join([os.path.expanduser('~'), 'mms_auth_info.pkl']), 'wb'))



import pyspedas

def rotmat_set_coords(varname:str, in_coords:str, out_coords:str):
    """ Set input and output coordinates on a rotation matrix

    Parameters
    ----------
    varname:str
        Variable name of the rotation matrix
    in_coords:str
        Input coordinates for the vectors to be rotated
    out_coords
        Output coordinates for the vectors to be rotated

    Returns
    -------
    None

    """

    md=pyspedas.get_data(varname, metadata=True)
    data_att = {'input_coord_sys':in_coords, 'output_coord_sys':out_coords}
    # We have a pointer to the internal attribute dictionary, no need to call store_data here
    md['data_att'] = data_att


import cdflib


def rfs_variables_to_load(files):
    """
    This function finds a list of variables to load
    from the RFS files (essentially the same behavior
    as the IDL code).
    """
    out = []
    if len(files) == 0:
        return []
    # the variables should be the same across all files
    file = files[0]

    new_cdflib = False
    if cdflib.__version__ > "0.4.9":
        new_cdflib = True
    else:
        new_cdflib = False


    cdf_file = cdflib.CDF(file)
    cdf_info = cdf_file.cdf_info()

    if new_cdflib:
        variables = cdf_info.rVariables + cdf_info.zVariables
    else:
        variables = cdf_info["rVariables"] + cdf_info["zVariables"]

    for variable in variables:
        if variable[0:7] != 'psp_fld':
            continue
        try:
            elements = cdf_file.varget(variable)
        except ValueError:
            continue
        if elements is None:
            continue
        if variable in out:
            continue
        out.append(variable)
    return out



import cdflib
import numpy as np

from pyspedas import tnames
from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim, zlim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror


from typing import List, Optional, Union


def isee_vlf(
    trange: List[str] = ["2017-03-30/12:00:00", "2017-03-30/15:00:00"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    cal_gain: bool = False,
    force_download: bool = False,
):
    """
    Load ISEE VLF data from ERG Science Center

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-30/12:00:00', '2017-03-30/15:00:00']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load. Valid values: 'ath', 'gak', 'hus', 'ist', 'kap', 'mam', 'nai', 'all'
            Default: ['all']

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    cal_gain: bool
        If True, use calibration parameters in the CDF file to apply a gain calibration.
        Default: False

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> vlf_vars=pyspedas.projects.erg.isee_vlf(trange=['2017-03-30/12:00:00', '2017-03-30/15:00:00'],site='ath')
    >>> print(vlf_vars)

    """

    site_code_all = ["ath", "gak", "hus", "ist", "kap", "mam", "nai"]

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = site_code_all

    site_code = list(set(site_code).intersection(site_code_all))

    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        prefix = "isee_vlf_" + site_input + "_"
        file_res = 3600.0
        pathformat = (
            "ground/vlf/"
            + site_input
            + "/%Y/%m/isee_vlf_"
            + site_input
            + "_%Y%m%d%H_v??.cdf"
        )

        loaded_data_temp = load(
            pathformat=pathformat,
            file_res=file_res,
            trange=trange,
            prefix=prefix,
            suffix=suffix,
            get_support_data=get_support_data,
            varformat=varformat,
            downloadonly=downloadonly,
            notplot=notplot,
            time_clip=time_clip,
            no_update=no_update,
            uname=uname,
            passwd=passwd,
            force_download=force_download,
        )

        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                # Most of the load routines use the last variable loaded to get the ROR metadata, but in this
                # case, it seems to want the first variable instead.  So we'll only pass that one.

                gatt = get_gatt_ror(downloadonly, [loaded_data[0]])
                print(
                    "**************************************************************************"
                )
                print(gatt["Logical_source_description"])
                print("")
                print(f'Information about {gatt["Station_code"]}')
                print(f'PI {gatt["PI_name"]}')
                print("")
                print(f'Affiliations: {gatt["PI_affiliation"]}')
                print("")
                print("Rules of the Road for ISEE VLF Data Use:")
                print("")
                for gatt_text in gatt["TEXT"]:
                    print(gatt_text)
                print(gatt["LINK_TEXT"])
                print(
                    "**************************************************************************"
                )
            except:
                print("printing PI info and rules of the road was failed")

        if (not downloadonly) and (not notplot):
            t_plot_name_list = list(
                set(
                    tnames([prefix + "ch1" + suffix, prefix + "ch2" + suffix])
                ).intersection(loaded_data)
            )
            options(t_plot_name_list, "zlog", 1)
            options(t_plot_name_list, "ytitle", "Frequency [Hz]")
            options(t_plot_name_list, "ysubtitle", "")
            if not cal_gain:
                options(t_plot_name_list, "ztitle", "V^2/Hz")
            else:
                print("Calibrating the gain of VLF antenna system...")
                file_name = get_data(t_plot_name_list[0], metadata=True)["CDF"][
                    "FILENAME"
                ]
                if isinstance(file_name, list):
                    file_name = file_name[0]
                cdf_file = cdflib.CDF(file_name)

                ffreq = cdf_file.varget("freq_vlf")
                gain_ch1 = cdf_file.varget("amplitude_cal_vlf_ch1")
                gain_ch2 = cdf_file.varget("amplitude_cal_vlf_ch2")

                gain_ch1_mod = np.interp(ffreq, gain_ch1[0], gain_ch1[1]) * 1.0e-9
                gain_ch2_mod = np.interp(ffreq, gain_ch2[0], gain_ch2[1]) * 1.0e-9

                t_plot_name = prefix + "ch1" + suffix
                tmp1 = get_data(t_plot_name)
                if tmp1 is not None:
                    tmp1_metadata = get_data(t_plot_name, metadata=True)
                    tmp1_y = tmp1[1] / gain_ch1_mod / gain_ch1_mod
                    store_data(
                        t_plot_name,
                        data={"x": tmp1[0], "y": tmp1_y, "v": tmp1[2]},
                        attr_dict=tmp1_metadata,
                    )

                t_plot_name = prefix + "ch2" + suffix
                tmp2 = get_data(t_plot_name)
                if tmp2 is not None:
                    tmp2_metadata = get_data(t_plot_name, metadata=True)
                    tmp2_y = tmp1[1] / gain_ch2_mod / gain_ch2_mod
                    store_data(
                        t_plot_name,
                        data={"x": tmp2[0], "y": tmp2_y, "v": tmp2[2]},
                        attr_dict=tmp2_metadata,
                    )

                options(t_plot_name_list, "ztitle", "nT^2/Hz")

    return loaded_data



import logging
import numpy as np
# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean
from pyspedas.tplot_tools import get_data, store_data, options, tnames


def mms_eis_spec_combine_sc(
        species='proton', data_units='flux', datatype='extof', data_rate='srvy',
        level='l2', suffix='',
    ):
    """
    Combines omni-directional energy spectrogram variable from EIS on multiple
        MMS spacecraft.

    Parameters
    ----------
        datatype: str
            'extof', 'electroenergy', or 'phxtof' (default: 'extof')

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst' (default: 'srvy')

        level: str
            data level ['l1a','l1b','l2pre','l2' (default)]

        data_units: str
            desired units for data, e.g., 'flux' or 'cps' (default: 'flux')

        suffix: str
            suffix of the loaded data; useful for preserving original tplot var

        species: str
            species for calculation, e.g., proton, oxygen, alpha or electron
            (default: 'proton')

    Returns
    --------
        Name of tplot variables created.
    """
    ## Thoughts for extensions:
    ## - Ensure arguments passed to modules are of lowecase

    if data_units == 'flux':
        units_label = 'Intensity\n[1/cm^2-sr-s-keV]'
    elif data_units == 'cps':
        units_label = 'CountRate\n[counts/s]'
    elif data_units == 'counts':
        units_label = 'Counts\n[counts]'

    #assert type(datatype) is str
    if not isinstance(species, list): species = [species]
    if not isinstance(datatype, list): datatype = [datatype]

    out_vars = []

    for species_id in species:
        for dtype in datatype:
            # retrieve: omni variables of species to determine # of probes
            _species = species_id
            if dtype == 'electronenergy':
                _species = 'electron'
            eis_sc_check = tnames('mms*eis*' + data_rate + '*' + dtype+'*' + _species + '*' + data_units + '*omni'+ suffix)

            # process multiple probes
            probes = []
            for name in eis_sc_check:
                probes.append(name[3:4])
            if len(probes) > 4:
                probes = probes[:-2]
            if len(probes) > 1:
                probe_string = probes[0] + '-' + probes[-1]
            else:
                if probes:
                    probe_string = probes[0]
                else:
                    logging.error('No probes found from eis_sc_check tnames.')
                    return

            allmms_prefix = 'mmsx_epd_eis_' + data_rate + '_' + level + '_' + dtype + '_'

            # DETERMINE SPACECRAFT WITH SMALLEST NUMBER OF TIME STEPS TO USE
            # AS A REFERENCE SPACECRAFT
            omni_vars = tnames('mms?_epd_eis_'+data_rate+'_'+level+'_'+dtype+'_'+_species+'_'+data_units+'_omni'+suffix)

            if not omni_vars:
                logging.error('No EIS '+dtype+'data loaded!')
                return

            time_size = np.zeros(len(probes))
            energy_size = np.zeros(len(probes))

            # Retrieve probe's pitch angle dist for all 6 (omni) telescopes
            for p, probe in enumerate(probes):
                # note: return from get_data here is (times, data, v)
                #       according to https://github.com/MAVENSDC/PyTplot/blob/ec87591521e84bae8d81caccaf64fc2a5785186f/pytplot/get_data.py#L66
                # note: there are also available 'spec_bins' values
                #print(pyspedas.tplot_tools.data_quants[omni_vars[p]].coords)
                #t, data, v = get_data(omni_vars[p])
                omni_times, omni_data, omni_energies = get_data(omni_vars[p])
                time_size[p] = len(omni_times)
                energy_size[p] = len(omni_energies)

            reftime_sc_loc = np.argmin(time_size)
            ref_sc_time_size = int(min(time_size))
            refenergy_sc_loc = np.argmin(energy_size)
            ref_sc_energy_size = int(min(energy_size))

            prefix = 'mms'+probes[reftime_sc_loc]+'_epd_eis_'+data_rate+'_'+level+'_'+dtype+'_'

            # Retrieve specific probe's data based on minimum time/energy
            # Note: I did not split these tuples as the namespace is reused, i.e., "_refprobe"
            time_refprobe = get_data(omni_vars[reftime_sc_loc])
            energy_refprobe = get_data(omni_vars[refenergy_sc_loc])

            # time x energy x spacecraft
            omni_spec_data = np.empty([len(time_refprobe[0]), len(energy_refprobe[2]), len(probes)])
            omni_spec_data[:] = np.nan
            # time x energy
            omni_spec = np.empty([len(time_refprobe[0]), len(energy_refprobe[2])])
            omni_spec[:] = np.nan

            energy_data = np.zeros([len(energy_refprobe[2]), len(probes)])
            common_energy = np.zeros(len(energy_refprobe[2]))

            # Average omni flux over all spacecraft and define common energy grid
            for pp in range(len(omni_vars)):
                temp_data = get_data(omni_vars[pp])
                energy_data[:,pp] = temp_data[2][0:len(common_energy)]
                omni_spec_data[0:ref_sc_time_size,:,pp] = temp_data[1][0:ref_sc_time_size,0:len(common_energy)]

            for ee in range(len(common_energy)):
                common_energy[ee] = nanmean(energy_data[ee,:], axis=0)

            # Average omni flux over all spacecraft
            for tt in range(len(time_refprobe[0])):
                for ee in range(len(energy_refprobe[2])):
                    omni_spec[tt,ee] = nanmean(omni_spec_data[tt,ee,:], axis=0)

            # store new tplot variable
            omni_spec[np.isnan(omni_spec)] = 0.
            new_name = allmms_prefix+_species+'_'+data_units+'_omni'
            store_data(new_name, data={'x':time_refprobe[0], 'y':omni_spec, 'v':energy_refprobe[2]})
            options(new_name, 'ylog', True)
            options(new_name, 'zlog', True)
            options(new_name, 'spec', True)
            options(new_name, 'ztitle', units_label)
            options(new_name, 'ytitle', ' \\ '.join(['mms'+probe_string, _species.upper(), 'Energy [keV]']))
            out_vars.append(new_name)

            # Spin-average the data
            spin_nums = get_data(prefix+'spin'+suffix)
            if spin_nums is None:
                logging.error('Error: Could not find EIS spin variable -- now ending procedure.')
                return

            # find where the spin starts
            _, spin_starts = np.unique(spin_nums[1], return_index=True)
            spin_sum_flux = np.zeros([len(spin_starts), len(omni_spec[0,:])])

            current_start = 0
            for spin_idx in range(len(spin_starts)):
                spin_sum_flux[spin_idx,:] = nanmean(omni_spec[current_start:spin_starts[spin_idx],:], axis=0)
                current_start = spin_starts[spin_idx] + 1

            sp = '_spin'
            new_name = allmms_prefix+_species+'_'+data_units+'_omni'+sp
            store_data(new_name, data={'x':spin_nums[0][spin_starts], 'y':spin_sum_flux, 'v':energy_refprobe[2]})
            options(new_name, 'spec', True)
            options(new_name, 'zlog', True)
            options(new_name, 'ylog', True)
            options(new_name, 'spec', True)
            out_vars.append(new_name)
    return out_vars




def initialize_list(the_list):
    """
    Recursively initializes a nested list by replacing each element with an empty list.

    Parameters:
        the_list (list): The list to be initialized.

    Returns:
        list: The initialized list.
    """
    index = 0
    for i in the_list:
        if hasattr(i, "__len__"):
            the_list[index] = initialize_list(i)
        else:
            the_list[index] = []
        index += 1
    return the_list


def place_values_in_list(the_list, location, to_append):
    """
    Appends a value to a list at the specified location.

    Parameters:
        the_list (list): The list to modify.
        location (int or list): The index or indices specifying the location in the list.
        to_append (any): The value to append to the list.

    Returns:
        None
    """
    testing = the_list
    if hasattr(location, "__len__"):
        for i in range(len(location)):
            testing = testing[location[i]]
        testing.append(to_append)
    else:
        testing = testing[location]
        testing.append(to_append)


def get_values_from_list(the_list, location):
    """
    Retrieves values from a nested list based on the provided location.

    Parameters:
        the_list (list): The nested list from which to retrieve values.
        location (int or list): The index or list of indices specifying the location of the desired values.

    Returns:
        The value(s) at the specified location in the nested list.
    """
    testing = the_list
    if hasattr(location, "__len__"):
        for i in range(len(location)):
            testing = testing[location[i]]
        return testing
    else:
        testing = testing[location]
        return testing




"""Test functions in the utilities folder."""
import unittest
import logging
from pyspedas import find_ip_address


class UtilTestCases(unittest.TestCase):
    """Test functions in the utilities folder."""

    def test_find_ip_address(self):
        """Show current public IP address """
        ip = find_ip_address()
        print("Current public IP address: %s" % ip)

if __name__ == '__main__':
    unittest.main()



import logging
from pyspedas.utilities.dailynames import dailynames, yearlynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from .config import CONFIG


def load(
    trange=["2013-11-5", "2013-11-6"],
    datatype="1min",
    level="hro",
    prefix="",
    suffix="",
    get_support_data=False,
    get_ignore_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=True,
    force_download=False,
):
    """
    Loads OMNI (Combined 1AU IP Data; Magnetic and Solar Indices) data; intended for internal use.

    This function is a core component for data loading but is not meant to be called directly by users.
    Instead, users should utilize the wrapper function `pyspedas.projects.omni.data` to access this functionality.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest specified as ['starttime', 'endtime'] with the format
        'YYYY-MM-DD' or 'YYYY-MM-DD/hh:mm:ss' to specify more or less than a day.
        default=['2013-11-5', '2013-11-6']
    datatype : str, optional
        Data type; valid options: '1min', '5min', 'hourly'.
        Default: '1min'.
    level : str, optional
        Data level; valid options: 'hro', 'hro2'. Default='hro'
    prefix : str, optional
        Prefix for the tplot variable names. By default, no prefix is added.
    suffix : str, optional
        Suffix for the tplot variable names. By default, no suffix is added.
    get_support_data : bool, optional
        If True, loads data with "VAR_TYPE" attribute value "support_data" into tplot.
        By default, only loads data with a "VAR_TYPE" attribute of "data".
    get_ignore_data : bool, optional
        If True, specific data types will be ignored during loading. This is especially
        used for hourly data where ignoring certain data is required.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character "*" is accepted.
        By default, all variables are loaded.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded.
    downloadonly : bool, optional
        If True, downloads the CDF files but does not load them into tplot variables.
    notplot : bool, optional
        If True, returns the data in hash tables instead of creating tplot variables.
    no_update : bool, optional
        If True, loads data only from the local cache.
    time_clip : bool, optional
        If True, clips the variables to exactly the range specified in the trange parameter.
    force_download: bool, optional
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------
    list or dict
        List of tplot variables created if notplot is False. Otherwise, returns a dictionary
        with data tables.

    Examples
    --------
    This function is not intended to be called directly.
    """

    file_res = 24 * 3600.0

    if "min" in datatype:
        pathformat = (
            level
            + "_"
            + datatype
            + "/%Y/omni_"
            + level
            + "_"
            + datatype
            + "_%Y%m01_v??.cdf"
        )
        remote_names = dailynames(file_format=pathformat, trange=trange, res=file_res)
    elif "hour" in datatype:
        pathformat = "hourly/%Y/omni2_h0_mrg1hr_%Y%m01_v??.cdf"
        get_ignore_data = True  # required to load these files
        remote_names = yearlynames(
            file_format=pathformat, trange=trange, resolution="half-year"
        )
    else:
        logging.error("Invalid datatype: " + datatype)
        return

    # find the full remote path names using the trange

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
        force_download=force_download,
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        prefix=prefix,
        suffix=suffix,
        get_support_data=get_support_data,
        get_ignore_data=get_ignore_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars



import logging
import pyspedas
import numpy as np
from pyspedas.tplot_tools import get_y_range
from pyspedas.tplot_tools import tplot_wildcard_expand


def options(name, option=None, value=None, opt_dict=None, quiet=False):
    """ Set a large variety of options for individual plots.

    Parameters
    ----------
        name : str or list[str]
            Names of tplot variables to be updated (wildcards accepted).
        option : str, optional
            The name of the option. See the options section below.
        value : str, int, float, list, optional
            The value of the option. See the options section below.
        opt_dict : dict, optional
            This can be a dictionary of option-value pairs. 'option' and 'value'
            will not be needed if this dictionary item is supplied.
        quiet: bool, optional
            If True, do not complain about unrecognized options.

    Options
    -------

    Many of the options are passed directly to matplotlib calls.  For more extensive documentation about how to use these
    obtions, see the matplotlib documentation: https://matplotlib.org/stable/users/index.html

    Note that many X-axis options are controlled at the level of the entire plot, rather than per-variable (since plots with multiple panels will
    share many X axis properties).  See the tplot_options() routine for available per-plot options,


        ======================  ===========  ===========================================================================================================================
        Panel Options           Value type   Notes
        ======================  ===========  ===========================================================================================================================
        title                   str          The title of the plot.
        panel_size              flt          Number between (0,1], representing the percent size of the plot.
        alpha                   flt          Number between [0,1], gives the transparency of the plot lines.
        line_width              flt          Sets plot line width.
        line_style              str          scatter (to make scatter plots), or solid_line, dot, dash, dash_dot, dash_dot_dot_dot, long_dash.
        border                  bool         Turns on or off the top/right axes that would create a box around the plot.
        var_label_format        str          The format of the tick labels if this variable is displayed as an alternative x axis. Default: {:.2f}
        char_size               int          Defines character size for x/y/z titles and subtitles
        right_axis              bool         If true,display a second Y axis on the right side of the plot.
        second_axis_size        numeric      The size of the second axis to display
        data_gap                numeric      If there is a gap in the data larger than this number in seconds, then insert
        (cont)                  (cont)       NaNs. This is similar to using the degap procedure on the variable, but is
        (cont)                  (cont)       applied at plot-time, and does not persist in the variable data.
        annotations             dict         A dictionary or list of dictionaries of matplotlib text annotation parameters (see annotate() routine)
        visible                 bool         If False, do not display lines for this variable.
        nodata                  bool         If True, do not display lines for this variable.
        ======================  ===========  ===========================================================================================================================

        ======================  ===========  ===========================================================================================================================
        Legend Options          Value type   Notes
        ======================  ===========  ===========================================================================================================================
        legend_names            list         A list of strings that will be used to identify the legends.
        legend_location         str          A string giving the location of the legend box.
        legend_size             numeric      The font size of the legend names
        legend_shadow           bool         Turns on or off drop shadows on the legend box
        legend_title            str          The title to display on the legend
        legend_titlesize        numeric      The font size of the legend title
        legend_color            [str]        The color of the legend names
        legend_edgecolor        str          The border color of the legend box
        legend_facecolor        str          The background color of the legend box
        legend_markerfirst      boolean      Put the marker and line to the left of the label in the legend
        legend_markerscale      numeric      The scale size of markers displayed in the legend
        legend_linewidth        numeric      The width of the lines displayed in the legend
        ======================  ===========  ===========================================================================================================================

        ======================  ===========  ===========================================================================================================================
        X Axis Options          Value type   Notes
        ======================  ===========  ===========================================================================================================================
        xtitle                  str          The title to be placed under the x axis.
        xsubtitle               str          The subtitle to be placed under the x axis
        xtitle_color            str          The color of the x axis title.
        xtick_length            numeric      The length of the x tick marks
        xtick_width             numeric      The width of the x tick marks
        xtick_color             str          The color of the x tick marks
        xtick_labelcolor        str          The color of the x tick marks
        xtick_direction         str          The direction of the x tick marks (in, out, inout)
        ======================  ===========  ===========================================================================================================================


        ======================  ===========  ===========================================================================================================================
        Y Axis Options          Value type   Notes
        ======================  ===========  ===========================================================================================================================
        y_range                 flt/list     Two numbers that give the y axis range of the plot. If a third argument is present, set linear or log scaling accordingly.
        ylog                    bool         True sets the y axis to log scale, False reverts.
        ytitle                  str          Title shown on the y axis. Use backslash for new lines.
        ysubtitle               str          Subtitle shown on the y axis.
        ytitle_color            str          The color of the y axis title.
        ytick_length            numeric      The length of the Y tick marks
        ytick_width             numeric      The width of the Y tick marks
        ytick_color             str          The color of the Y tick marks
        ytick_labelcolor        str          The color of the Y tick marks
        ytick_direction         str          The direction of the Y tick marks (in, out, inout)
        y_major_ticks           [numeric]    A list of values that will be used to set the major ticks on the Y axis.
        y_minor_tick_interval   numeric      The interval between minor ticks on the Y axis.
        ======================  ===========  ===========================================================================================================================

        ======================  ===========  ===========================================================================================================================
        Error Bar Options       Value type   Notes
        ======================  ===========  ===========================================================================================================================
        errorevery              numeric      Interval at which to show error bars
        capsize                 numeric      The size of error bar caps
        ecolor                  str          The color of the error bar lines
        elinewidth              numeric      The width of the error bar lines
        ======================  ===========  ===========================================================================================================================

        ======================  ===========  ===========================================================================================================================
        Marker/Symbol  Options  Value type   Notes
        ======================  ===========  ===========================================================================================================================
        marker_size             numeric      The size of the markers displayed in the plot
        markevery               numeric      Interval at which to show markers
        symbols                 bool         If True, display as a scatter plot (no lines)
        ======================  ===========  ===========================================================================================================================


        ======================  ===========  ===========================================================================================================================
        Z  / Specplot Options   Value type   Notes
        ======================  ===========  ===========================================================================================================================
        spec                    bool         Display this variable as a spectrogram.
        colormap                str/list     Color map to use for specplots https://matplotlib.org/examples/color/colormaps_reference.html.
        z_range                 flt/list     Two numbers that give the z axis range of the plot. If a third argument is present, set linear or log scaling accordingly.
        zlog                    int          True sets the z axis to log scale, False reverts (spectrograms only).
        ztitle                  str          Title shown on the z axis. Spec plots only. Use backslash for new lines.
        zsubtitle               str          Subtitle shown on the z axis. Spec plots only.
        ztitle_color            str          The color of the z axis title.
        x_interp                bool         If true, perform smoothing of spectrograms in the X direction
        x_interp_points         numeric      Number of interpolation points to use in the X direction
        y_interp                bool         If true, perform smoothing of spectrograms in the Y direction
        y_interp_points         numeric      Number of interpolation points to use in the Y direction
        xrange_slice            flt/list     Two numbers that give the x axis range of spectrogram slicing plots.
        yrange_slice            flt/list     Two numbers that give the y axis range of spectrogram slicing plots.
        xlog_slice              bool         Sets x axis on slice plot to log scale if True.
        ylog_slice              bool         Sets y axis on slice plot to log scale if True.
        spec_dim_to_plot        int/str      If variable has more than two dimensions, this sets which dimension the "v"
        (cont)                  (cont)       variable will display on the y axis in spectrogram plots.
        (cont)                  (cont)       All other dimensions are summed into this one, unless "spec_slices_to_use"
        (cont)                  (cont)       is also set for this variable.
        spec_slices_to_use      str          Must be a dictionary of coordinate:values. If a variable has more than two
        (cont)                  (cont)       dimensions, spectrogram plots will plot values at that particular slice of
        (cont)                  (cont)       that dimension. See examples for how it works.
        ======================  ===========  ===========================================================================================================================

        Many options have synonyms or variant spellings that are commonly used.  The first column gives the name that is used
        throughout the plotting code.  The second column gives the synonyms that are accepted.

        ======================  ======================================================================================================================================
        Canonical name          Accepted synonyms
        ======================  ======================================================================================================================================
        title                   name
        line_color              color, colors, col, cols, line_colors
        legend_names            labels, legend_name, legend_label, legend_labels
        legend_location         labels_location, legends_location, label_location, labels_location
        legend_size             labels_size, label_size
        legend_shadow           labels_shadow. label_shadow
        legend_title            label_title, labels_title
        legend_titlesize        lable_titlesize, labels_titlesize
        legend_color            legends_color, label_color, labels_color
        legend_edgecolor        label_edgecolor, labels_edgecolor
        legend_facecolor        label_facecolor, labels_facecolor
        legend_markerfirst      label_markerfirst, labels_markerfirst
        legend_markerscale      label_markerscale, labels_markerscale
        legend_linewidth        label_linewidth, labels_linewidth
        legend_frameon          label_frameon, labels_frameon
        legend_ncols            label_ncols, labels_ncols
        line_style_name         line_style, linestyle
        char_size               charsize
        marker                  markers
        marker_size             markersize
        markevery               markerevery mark_every marker_every
        symbol                  symbols
        line_width              thick
        y_range                 yrange
        z_range                 zrange
        data_gap                datagap
        spec_dim_to_plot        spec_plot_dim
        var_label_format        varlabel_format
        annotations             annotation
        ======================  ======================================================================================================================================



    Returns
    -------
        None

    Examples
    --------
        >>> # Change the y range of Variable1
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> pyspedas.options('Variable1', 'yrange', [2,4])

        >>> # Change Variable1 to use a log scale
        >>> pyspedas.options('Variable1', 'ylog', 1)
        >>> pyspedas.tplot('Variable1')

        >>> # Multi-dimensional variable
        >>> y_data = np.random.rand(5, 4, 3)
        >>> v1_data = [0, 1, 3, 4]
        >>> v2_data = [1, 2, 3]
        >>> pyspedas.store_data("Variable2", data={'x': x_data, 'y': y_data, 'v1': v1_data, 'v2': v2_data})
        >>> # Set the spectrogram plots to show dimension 'v2' at slice 'v1' = 0
        >>> pyspedas.options('Variable2', 'spec', 1)
        >>> pyspedas.options("Variable2", "spec_dim_to_plot", 'v2')
        >>> pyspedas.options("Variable2", "spec_slices_to_use", {'v1': 0})
        >>> pyspedas.tplot('Variable2')

    """

    if isinstance(name, int):
        name = list(pyspedas.tplot_tools.data_quants.keys())[name]

    if opt_dict is None:
        opt_dict = {option: value}
    else:
        if not isinstance(opt_dict,dict):
            logging.error("dict must be a dictionary object.  Returning.")
            return

    if not isinstance(name, list):
        name = [name]

    names = tplot_wildcard_expand(name)
    for i in names:

        for option, value in opt_dict.items():

            # Lower case option for consistency
            option = option.lower()

            if i not in pyspedas.tplot_tools.data_quants.keys():
                logging.info(str(i) + " is currently not in pyspedas.")
                return

            elif option in ['color', 'colors', 'col', 'cols', 'line_color', 'line_colors']:
                if isinstance(value, list):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['line_color'] = value
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['line_color'] = [value]

            elif option == 'link':
                if isinstance(value, list):
                    pyspedas.link(i, value[1], value[0])

            elif option in ['annotation', 'annotations']:
                # It is probably more convenient to use the annotations() wrapper function
                # to manage annotations
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['annotations'] = value

            elif option == 'colormap':
                if isinstance(value, list):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['colormap'] = value
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['colormap'] = [value]

            elif option == 'second_axis_size':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['second_axis_size'] = value

            elif option == 'spec':
                _reset_plots(i)
                if value:
                    if 'spec_bins' not in pyspedas.tplot_tools.data_quants[i].coords:
                        logging.warning(f"{i} does not contain coordinates for spectrogram plotting.  Continuing...")
                        continue
                    else:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec'] = value
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_range'] = get_y_range(pyspedas.tplot_tools.data_quants[i])

                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec'] = value
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_range'] = get_y_range(pyspedas.tplot_tools.data_quants[i])

                # Set the default dimension to plot by.  All others will be summed over.
                if 'spec_dim_to_plot' not in pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']:
                    if 'v' in pyspedas.tplot_tools.data_quants[i].coords:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_dim_to_plot'] = 'v'
                    elif 'v2' in pyspedas.tplot_tools.data_quants[i].coords:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_dim_to_plot'] = 'v2'
                    else:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_dim_to_plot'] = 'v1'

            elif option in ['legend_names', 'labels', 'legend_name', 'legend_label', 'legend_labels']:
                if isinstance(value, list):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_names'] = value
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_names'] = [value]

            elif option in ['legend_location', 'labels_location', 'legends_location', 'label_location', 'labels_location']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_location'] = value

            elif option in ['legend_size', 'labels_size', 'label_size']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_size'] = value

            elif option in ['legend_shadow', 'labels_shadow', 'label_shadow']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_shadow'] = value

            elif option in ['legend_title', 'label_title', 'labels_title']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_title'] = value

            elif option in ['legend_titlesize', 'label_titlesize', 'labels_titlesize']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_titlesize'] = value

            elif option in ['legend_color', 'legends_color','label_color','labels_color']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_color'] = value

            elif option in ['legend_edgecolor', 'label_edgecolor', 'labels_edgecolor']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_edgecolor'] = value

            elif option in ['legend_facecolor', 'label_facecolor', 'labels_facecolor']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_facecolor'] = value

            elif option in ['legend_markerfirst', 'label_markerfirst', 'labels_markerfirst']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_markerfirst'] = value

            elif option in ['legend_markerscale', 'label_markerscale', 'labels_markerscale']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_markerscale'] = value

            elif option in ['legend_linewidth', 'label_linewidth', 'labels_linewidth']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_linewidth'] = value

            elif option in ['legend_frameon', 'label_frameon', 'labels_frameon']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_frameon'] = value

            elif option in ['legend_ncols', 'labels_ncols', 'label_ncols']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['legend_ncols'] = value

            elif option == 'xlog_slice':
                if value:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['slice_xaxis_opt']['xi_axis_type'] = 'log'
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['slice_xaxis_opt']['xi_axis_type'] = 'linear'

            elif option == 'ylog':
                negflag = 0 # _ylog_check(data_quants, value, i)
                if negflag == 0 and value:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_axis_type'] = 'log'
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_axis_type'] = 'linear'

            elif option == 'ylog_slice':
                if value:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['slice_yaxis_opt']['yi_axis_type'] = 'log'
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['slice_yaxis_opt']['yi_axis_type'] = 'linear'

            elif option == 'zlog':
                # check for negative values and warn the user that they will be ignored
                negflag = _zlog_check(pyspedas.tplot_tools.data_quants, value, i)
                if negflag != 0 and value:
                    logging.warning(str(i) + ' contains negative values; setting the z-axis to log scale will cause the negative values to be ignored on figures.')

                if value:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['z_axis_type'] = 'log'
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['z_axis_type'] = 'linear'

            elif option =='visible':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['visible'] = bool(value)

            elif option =='nodata':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['visible'] = not bool(value)

            # Obsolete? (except for value='none'?) JWL 2024-03-21
            # These don't seem to be the correct format for matplotlib parameterized line styles.
            elif option in ['line_style', 'linestyle', 'line_style_name']:
                if value == 0 or value == 'solid_line':
                    to_be = []
                elif value == 1 or value == 'dot':
                    to_be = [2, 4]
                elif value == 2 or value == 'dash':
                    to_be = [6]
                elif value == 3 or value == 'dash_dot':
                    to_be = [6, 4, 2, 4]
                elif value == 4 or value == 'dash_dot_dot_dot':
                    to_be = [6, 4, 2, 4, 2, 4, 2, 4]
                elif value == 5 or value == 'long_dash':
                    to_be = [10]
                else:
                    to_be=value

                # This does not appear to be used by tplot. JWL 2024-03-21
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['line_style'] = to_be

                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['line_style_name'] = _convert_to_matplotlib_linestyle(value)

                # unused?
                if(value == 6 or value == 'none'):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['visible'] = False

            elif option in ['char_size', 'charsize']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['char_size'] = value

            elif option in ['var_label_format', 'varlabel_format']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['var_label_format'] = value

            elif option in ['name', 'title']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['title'] = value

            elif option == "panel_size":
                if value > 1 or value <= 0:
                    logging.info("Invalid panel_size value (%f). Should be in (0, 1]",value)
                    return
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['panel_size'] = value

            elif option == 'alpha':
                if value > 1 or value < 0:
                    logging.info("Invalid alpha value (%f). Should be [0, 1]",value)
                    return
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['alpha'] = value

            elif option in ['marker', 'markers']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['marker'] = value

            elif option == 'errorevery':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['errorevery'] = value

            elif option == 'capsize':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['capsize'] = value

            elif option == 'ecolor':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['ecolor'] = value

            elif option == 'elinewidth':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['elinewidth'] = value

            elif option in ['marker_size', 'markersize']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['marker_size'] = value

            elif option in ['markevery', 'markerevery', 'mark_every', 'marker_every']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['markevery'] = value

            elif option in ['symbols', 'symbol']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['symbols'] = value

            elif option == 'xtick_length':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['xtick_length'] = value

            elif option == 'ytick_length':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['ytick_length'] = value

            elif option == 'xtick_width':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['xtick_width'] = value

            elif option == 'ytick_width':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['ytick_width'] = value

            elif option == 'xtick_color':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['xtickcolor'] = value

            elif option == 'ytick_color':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['ytickcolor'] = value

            elif option == 'xtick_labelcolor':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['xtick_labelcolor'] = value

            elif option == 'ytick_labelcolor':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['ytick_labelcolor'] = value

            elif option == 'xtick_direction':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['xtick_direction'] = value

            elif option == 'ytick_direction':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['ytick_direction'] = value

            elif option == 'right_axis':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['right_axis'] = value

            elif option in ['thick', 'line_width']:
                if isinstance(value, list):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['line_width'] = value
                else:
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['line_opt']['line_width'] = [value]

            elif option in ['yrange', 'y_range']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_range'] = [value[0], value[1]]
                # track whether the yrange option was set by the user
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_range_user'] = True
                # IDL SPEDAS supports a 3-argument form of yrange, where the third argument selects
                # log or linear scaling
                if len(value) == 3:
                    if value[2]:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_axis_type'] = 'log'
                    else:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_axis_type'] = 'linear'

            elif option == 'y_major_ticks':
                # check whether the value is 1D array-like
                if isinstance(value, (list, np.ndarray)):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_major_ticks'] = value
                else:
                    logging.warning('y_major_ticks must be a 1D array-like object')

            elif option == 'y_minor_tick_interval':
                # check whether the value is a number
                if isinstance(value, (int, float)):
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_minor_tick_interval'] = value
                else:
                    logging.warning('y_minor_tick_interval must be a number')

            elif option in ['zrange', 'z_range']:
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['z_range'] = [value[0], value[1]]
                if len(value) == 3:
                    # IDL SPEDAS supports a 3-argument form of zrange, where the third element specifies linear vs. log scaling.
                    negflag = _zlog_check(pyspedas.tplot_tools.data_quants, value[2], i)
                    if negflag != 0 and value:
                        logging.warning(str(i) + ' contains negative values; setting the z-axis to log scale will cause the negative values to be ignored on figures.')

                    if value[2]:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['z_axis_type'] = 'log'
                    else:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['z_axis_type'] = 'linear'

            elif option == 'xrange_slice':
                plt_opts = pyspedas.tplot_tools.data_quants[i].attrs['plot_options']
                if plt_opts.get('slice_xaxis_opt') is not None:
                    plt_opts['slice_xaxis_opt']['xi_range'] = [value[0], value[1]]

            elif option == 'yrange_slice':
                plt_opts = pyspedas.tplot_tools.data_quants[i].attrs['plot_options']
                if plt_opts.get('slice_yaxis_opt') is not None:
                    plt_opts['slice_yaxis_opt']['yi_range'] = [value[0], value[1]]

            elif option == 'xtitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['xaxis_opt']['axis_label'] = value

            elif option == 'ytitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['axis_label'] = value

            elif option == 'ztitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['axis_label'] = value

            elif option == 'xsubtitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['xaxis_opt']['axis_subtitle'] = value

            elif option == 'ysubtitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['axis_subtitle'] = value

            elif option == 'zsubtitle':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['axis_subtitle'] = value

            elif option == 'xtitle_color':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['xaxis_opt']['axis_color'] = value

            elif option == 'ytitle_color':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['axis_color'] = value

            elif option == 'ztitle_color':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['zaxis_opt']['axis_color'] = value

            elif option in ['data_gap', 'datagap']: #jmm, 2023-06-20
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['data_gap'] = value

            elif option in ['spec_dim_to_plot', 'spec_plot_dim']:
                if len(pyspedas.tplot_tools.data_quants[i].values.shape) <= 2:
                    logging.warning(f"Must have more than 2 coordinate dimensions to set spec_coord_to_plot for {pyspedas.tplot_tools.data_quants[i].name}")
                    continue

                # Set the 'spec_dim_to_plot' value to either 'v' or 'v1', 'v2', 'v3', etc.
                if isinstance(value, int):
                    coord_to_plot = "v" + str(value)
                    if coord_to_plot not in pyspedas.tplot_tools.data_quants[i].coords:
                        if value == 1:
                            coord_to_plot = "v"
                            if coord_to_plot not in pyspedas.tplot_tools.data_quants[i].coords:
                                logging.warning(f"Dimension {value} not found in {pyspedas.tplot_tools.data_quants[i].name}")
                                continue
                        else:
                            logging.warning(f"Dimension {value} not found in {pyspedas.tplot_tools.data_quants[i].name}")
                            continue
                    pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_dim_to_plot'] = coord_to_plot
                elif isinstance(value, str):
                    coord_to_plot = value
                    if coord_to_plot not in pyspedas.tplot_tools.data_quants[i].coords:
                        logging.warning(f"Dimension {value} not found in {pyspedas.tplot_tools.data_quants[i].name}")
                        continue
                    else:
                        pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_dim_to_plot'] = value

                # If we're plotting against different coordinates, we need to change what we consider the "spec_bins"
                pyspedas.tplot_tools.data_quants[i].coords['spec_bins'] = pyspedas.tplot_tools.data_quants[i].coords[coord_to_plot]
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_range'] = get_y_range(pyspedas.tplot_tools.data_quants[i])

            elif option == 'spec_slices_to_use':
                if not isinstance(value, dict):
                    logging.error("Must be a dictionary object in the format {'v2':15, 'v3':7}")
                    return
                else:
                    for coord in value:
                        if coord not in pyspedas.tplot_tools.data_quants[i].coords:
                            logging.warning(f"Dimension {coord} not found in {pyspedas.tplot_tools.data_quants[i].name}")
                            continue

                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['spec_slices_to_use'] = value

            elif option == 'border':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['extras']['border'] = value

            elif option == 'y_interp':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_interp'] = value

            elif option == 'y_interp_points':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_interp_points'] = value

            elif option == 'x_interp':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['x_interp'] = value

            elif option == 'x_interp_points':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['x_interp_points'] = value
            elif option == 'y_no_resample':
                pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['yaxis_opt']['y_no_resample'] = value

            else:
                # Apparently cdf_to_tplot is treating all variable attributes as potential plot
                # options.  Adding this warning will end up spamming the logs unless cdf_to_tplot is changed.
                if not quiet:
                    logging.warning(f"Unrecognized option {option}")
                #pass
    return


def _ylog_check(data_quants, value, i):
    negflag = 0
    namedata = data_quants[i]
    # check variable data
    # if negative numbers, don't allow log setting
    datasets = [namedata]
    for oplot_name in namedata.attrs['plot_options']['overplots']:
        datasets.append(data_quants[oplot_name])

    if value == 1:
        for dataset in datasets:
            if 'spec' not in dataset.attrs['plot_options']['extras']:
                if dataset.min(skipna=True) < 0:
                    logging.warning('Negative data is incompatible with log plotting.')
                    negflag = 1
                    break
            else:
                if dataset.attrs['plot_options']['extras']['spec'] == 1:
                    if dataset.coords['spec_bins'].min(skipna=True) < 0:
                        logging.warning('Negative data is incompatible with log plotting.')
                        negflag = 1
                        break
    elif value != 1:
        # Using the 'negflag' as a way to not log something if the user doesn't want it to be logged
        negflag = 1
    return negflag


def _zlog_check(data_quants, value, i):
    negflag = 0
    namedata = data_quants[i]
    # check variable data
    # if negative numbers, don't allow log setting
    datasets = [namedata]
    for oplot_name in namedata.attrs['plot_options']['overplots']:
        datasets.append(data_quants[oplot_name])

    for dataset in datasets:
        if value == 1:
            if 'spec' in dataset.attrs['plot_options']['extras']:
                if dataset.attrs['plot_options']['extras']['spec'] == 1:
                    if dataset.min(skipna=True) < 0:
                        negflag = 1
                        break
        elif value != 1:
            # Using the 'negflag' as a way to not log something if the user doesn't want it to be logged
            negflag = 1
    return negflag


def _reset_plots(name):
    if isinstance(pyspedas.tplot_tools.data_quants[name], dict):  # non-record varying variable
        return
    pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['spec'] = 0
    pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['alt'] = 0
    pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['map'] = 0
    pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras']['plotter'] = None

def _convert_to_matplotlib_linestyle(linestyle):
    if not isinstance(linestyle,list):
        linestyle = [linestyle]
    converted_linestyles = []
    for ls in linestyle:
        if ls == 'solid_line':
            converted_linestyles.append('solid')
        elif ls == 'dot':
            converted_linestyles.append('dotted')
        elif ls == 'dash':
            converted_linestyles.append('dashed')
        elif ls == 'dash_dot':
            converted_linestyles.append('dashdot')
        else:
            converted_linestyles.append(ls)
    return converted_linestyles





from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from typing import List, Union
from .config import CONFIG

def load(trange:List[str]=['2009-01-01', '2009-01-02'],
         instrument:str='vhm',
         datatype:str='1min',
         prefix:str='',
         suffix:str='',
         get_support_data:bool=False,
         varformat:str=None,
         varnames:List[str]=[],
         downloadonly:bool=False,
         force_download:bool=False,
         notplot:bool=False,
         no_update:bool=False,
         time_clip:bool=False) -> List[str]:
    """
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2009-01-01', '2009-01-02']

        instrument: str
            Instrument to load.
            Default: 'vhm'

        datatype: str
            Data type
            Default: '1min'

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: '' (all variables loaded)

        varnames: list of str
            List of variable names to load (If empty list or not specified,
            all data variables are loaded)
            Default: []

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        force_download: bool
            Set this flag to download the CDF files, even if the local file is newer
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    -------
        list of str
            List of tplot variables created.

    Notes
    _____
    This function is not meant to be called directly; please see the instrument specific wrappers in __init__.py
    """

    if prefix is None:
        prefix=''
    if suffix is None:
        suffix=''

    if instrument == 'vhm':
        pathformat = 'mag_cdaweb/vhm_'+datatype+'/%Y/uy_'+datatype+'_vhm_%Y%m%d_v??.cdf'
    elif instrument == 'swoops':
        if datatype in ['bai_m0', 'bai_m1', 'bae_m0']:
            pathformat = 'plasma/swoops_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
        else:
            pathformat = 'plasma/swoops_cdaweb/'+datatype+'/%Y/uy_'+datatype+'_%Y0101_v??.cdf'
    elif instrument == 'swics':
        pathformat = 'plasma/swics_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'urap':
        pathformat = 'radio/urap_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'epac':
        if datatype == 'epac_m1':
            pathformat = 'particle/epac_cdaweb/'+datatype+'/%Y/uy_m1_epa_%Y%m%d_v??.cdf'
    elif instrument == 'hiscale':
        pathformat = 'particle/hiscale_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'
    elif instrument == 'cospin':
        pathformat = 'particle/cospin_cdaweb/'+datatype+'/%Y/uy_m0_'+datatype+'_%Y%m%d_v??.cdf'
    elif instrument == 'grb':
        pathformat = 'gamma/grb_cdaweb/'+datatype+'/%Y/uy_'+datatype.split('_')[1]+'_'+datatype.split('_')[0]+'_%Y%m%d_v??.cdf'


    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import os

CONFIG = {
    "local_data_dir": "fast_data/",
    "remote_data_dir": "https://spdf.gsfc.nasa.gov/pub/data/fast/",
}

# override local data directory with environment variables
if os.environ.get("FAST_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["FAST_DATA_DIR"]
elif os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "fast"])



from .load import load
from pyspedas.utilities.datasets import find_datasets

# This routine was originally in stereo/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def datasets(instrument=None, label=True):
    return find_datasets(mission='STEREO', instrument=instrument, label=label)



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
import datetime


def timestamp(val):
    """
    This function will turn on a time stamp that shows up at the bottom of every generated plot.

    Parameters
    ----------
        val  str
            A string that can either be 'on' or 'off'.

    Returns
    -------
        None

    Examples
    --------
         >>> # Turn on the timestamp
         >>> import pyspedas
         >>> pyspedas.timestamp('on')
    """

    if val == 'on':
        todaystring = datetime.datetime.now().strftime('%Y-%m-%d %H%M%S')
        pyspedas.tplot_tools.extra_layouts['time_stamp'] = todaystring
    else:
        if 'time_stamp' in pyspedas.tplot_tools.extra_layouts:
            del pyspedas.tplot_tools.extra_layouts['time_stamp']

    return



"""
Coordinate transformations.

Transform from/to the following coordinate systems:
GSE, GSM, SM, GEI, GEO, MAG, J2000

Times are in Unix seconds for consistency.

Notes
-----
This function is similar to cotrans.pro of IDL SPEDAS.
"""

import logging
import numpy as np
import pyspedas
from pyspedas import (
    store_data,
    get_data,
    get_coords,
    set_coords,
    tplot_copy,
)
from pyspedas.cotrans_tools.cotrans_lib import subcotrans


def cotrans(
    name_in=None,
    name_out=None,
    time_in=None,
    data_in=None,
    coord_in=None,
    coord_out=None,
):
    """
    Transform data from coord_in to coord_out.

    Parameters
    ----------
    name_in: str, optional
        Tplot name for input data.
    name_out: str, optional
        Tplot name for output data.
    time_in: list of float, optional
        Time array.
        Ignored if name_in is provided.
    data_in: list of float, optional
        Data in the coord_in system.
        Ignored if name_in is provided.
    coord_in: str
        Name of input coordinate system.
        Valid options: "gse", "gsm", "sm", "gei", "geo", "mag", "j2000"
    coord_out: str
        Name of output coordinate system.
        Valid options: "gse", "gsm", "sm", "gei", "geo", "mag", "j2000"

    Returns
    -------
    Returns 1 for successful completion, 0 for error.
        Fills a new tplot variable with data in the coord_out system.

    np.array
        If name_in and name_out are not provided, returns the transformed data array.
    """
    if coord_out is None:
        logging.error("cotrans error: No output coordinates were provided.")
        return 0

    # Input data may be specified as a bare array rather than a tplot variable
    if name_in is not None:
        var_coord_in = get_coords(name_in)
    else:
        var_coord_in = None

    # If the input coordinate system is supplied as an argument, and the tplot variable has a coordinate system
    # specified in its metadata, check that they match, and if not, log the error and return failure.

    if var_coord_in is not None and coord_in is not None:
        if var_coord_in.lower() != coord_in.lower():
            logging.error(
                f"cotrans error: {name_in} has {var_coord_in.lower()} coordinates, but transform from {coord_in.lower()} was requested."
            )
            return 0

    if coord_in is None:
        coord_in = var_coord_in
        if coord_in is None:
            logging.error("cotrans error: No input coordinates were provided.")
            return 0

    coord_in = coord_in.lower()
    coord_out = coord_out.lower()
    all_coords = ["gse", "gsm", "sm", "gei", "geo", "mag", "j2000"]

    if coord_in not in all_coords:
        logging.error(
            f"cotrans error: Requested input coordinate system {coord_in} not supported."
        )
        return 0
    if coord_out not in all_coords:
        logging.error(
            f"cotrans error: Requested output coordinate system {coord_out} not supported.",
        )
        return 0

    if name_in is not None:
        # If a name_in is provided, use it for data.
        tplot_data = get_data(name_in)
        time_in = tplot_data[0]
        data_in = tplot_data[1]

    if len(data_in[:]) < 1:
        logging.error("cotrans error: Data is empty.")
        return 0

    # Perform coordinate transformation.
    if not isinstance(data_in, np.ndarray) and not isinstance(data_in, list):
        data_in = np.array(data_in)
    if not isinstance(time_in, np.ndarray) and not isinstance(time_in, list):
        time_in = np.array(time_in)

    data_out = subcotrans(time_in, data_in, coord_in, coord_out)

    # If no input or output tplot variable names are provided, return the transformed data.
    if name_in is None and name_out is None:
        return data_out

    # Otherwise, store the data in a tplot variable.
    if name_in is None:
        store_data("cotranstemp", data={"x": time_in, "y": data_in})
        name_in = "cotranstemp"

    # Find the name of the output tplot variable.
    if name_out is None:
        name_out = name_in + "_" + coord_out

    # Save output tplot variable.
    tplot_copy(name_in, name_out)
    pyspedas.tplot_tools.data_quants[name_out].data = data_out

    # We should change an attribute for the coordinate system.
    set_coords(name_out, coord_out.upper())

    # Code to update the legend and axis labels has been moved into cotrans_set_coord().

    logging.info("Output variable: " + name_out)

    return 1



import numpy as np

from copy import deepcopy

from pyspedas import tinterpol, tcopy
from pyspedas.tplot_tools import time_double, tnames
from pyspedas.tplot_tools import time_string

from pyspedas.particles.moments.spd_pgs_moments import spd_pgs_moments
from pyspedas.particles.spd_part_products.spd_pgs_regrid import spd_pgs_regrid
from pyspedas import get_timespan, get_data, store_data, ylim

from .erg_lepe_get_dist import erg_lepe_get_dist
from .erg_lepi_get_dist import erg_lepi_get_dist
from .erg_pgs_clean_data import erg_pgs_clean_data
from .erg_pgs_limit_range import erg_pgs_limit_range
from .erg_convert_flux_units import erg_convert_flux_units
from .erg_pgs_moments_tplot import erg_pgs_moments_tplot
from .erg_pgs_make_fac import erg_pgs_make_fac
from .erg_pgs_make_e_spec import erg_pgs_make_e_spec
from .erg_pgs_make_theta_spec import erg_pgs_make_theta_spec
from .erg_pgs_make_phi_spec import erg_pgs_make_phi_spec
from .erg_pgs_do_fac import erg_pgs_do_fac
from .erg_pgs_progress_update import erg_pgs_progress_update
from .erg_pgs_make_tplot import erg_pgs_make_tplot

def erg_lep_part_products(
    in_tvarname,
    species=None,
    outputs=['energy'],
    no_ang_weighting=True,
    suffix='',
    units='flux',
    datagap=32.1,
    regrid=[16, 16],
    pitch=[0., 180.],
    theta=[-90., 90.],
    phi_in=[0., 360.],
    gyro=[0., 360.],
    energy=None,
    fac_type='mphism',
    trange=None,
    mag_name=None,
    pos_name=None,
    relativistic=False,
    no_regrid=False
    ):

    if len(tnames(in_tvarname)) < 1:
        print('No input data, please specify tplot variable!')
        return 0

    in_tvarname = tnames(in_tvarname)[0]
    vn_info = in_tvarname.split('_')
    instnm = vn_info[1]  #  ;; should be 'lepe' or 'lepi'
    lvl = vn_info[2]  #  ;; 'l2'
    datnm = vn_info[3]  #  ;; '3dflux'

    #  ;; Check if a tplot variable given is acceptable
    if (instnm != 'lepe') and (instnm != 'lepi') or (datnm != '3dflux'):
        print('The tplot variable given as an argument is not valid!')
        print(f'varname: {in_tvarname}')
        return

    if no_ang_weighting:
        no_regrid = True

    if isinstance(outputs, str):
        outputs_lc = outputs.lower()
        outputs_lc = outputs_lc.split(' ')
    elif isinstance(outputs, list):
        outputs_lc = []
        for output in outputs:
            outputs_lc.append(output.lower())

    units_lc = units.lower()

    if phi_in != [0., 360.]:
        if abs(phi_in[1] - phi_in[0]) > 360.:
            print('ERROR: Phi restrictions must have range no larger than 360 deg')
            return 0
        
        phi = phi_in  # Survey or implement of spd_pgs_map_azimuth() have not conducted yet.
        if phi[0] == phi[1]:
            phi = [0., 360.]


    if abs(gyro[1] - gyro[0]) > 360.:
        print('ERROR: Gyro restrictions must have range no larger than 360 deg')
        return 0
    if gyro[0] == gyro[1]:
        gyro = [0., 360.]


    """
    ;;Create energy spectrogram after FAC transformation if limits are not 
    ;;identical to the default.
    """
    if (gyro != [0., 360.]) or (pitch != [0.,180.]):
        idx = np.where(np.array(outputs_lc) == 'energy')[0]
        if (idx.shape[0] > 0) and ('fac_energy' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_energy'

        idx = np.where(np.array(outputs_lc) == 'moments')[0]
        if (idx.shape[0] > 0) and ('fac_moments' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_moments'

    #  ;;Preserve the original time range
    tr_org = get_timespan(in_tvarname)

    if instnm == 'lepe':
        times_array = erg_lepe_get_dist(in_tvarname, species=species, units=units_lc, time_only=True)
    if instnm == 'lepi':
        times_array = erg_lepi_get_dist(in_tvarname, species=species, units=units_lc, time_only=True)

    if trange is not None:
        
        trange_double = time_double(trange)
        time_indices = np.where((times_array >= trange_double[0]) \
                                & (times_array <= trange_double[1]))[0]
        
        if time_indices.shape[0] < 1:
            print(f'No ,{in_tvarname}, data for time range ,{time_string(trange_double)}')

    elif trange is None:
        time_indices = np.arange(times_array.shape[0])

    times_array = times_array[time_indices]



    if instnm == 'lepe':
        dist_all_time_range =  erg_lepe_get_dist(in_tvarname, time_indices, species=species, units=units_lc)
        dist = deepcopy(dist_all_time_range)
    if instnm == 'lepi':
        dist = erg_lepi_get_dist(in_tvarname, 0, species=species, units=units_lc)

    if 'energy' in outputs_lc:
        out_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))
    if 'theta' in outputs_lc:
        if instnm == 'lepe':
            n_theta_unique = len(np.unique(dist['theta']))
            out_theta = np.zeros((times_array.shape[0], n_theta_unique))
            out_theta_y = np.zeros((times_array.shape[0], n_theta_unique))
        elif  instnm == 'lepi':
            out_theta = np.zeros((times_array.shape[0], dist['n_theta']))
            out_theta_y = np.zeros((times_array.shape[0], dist['n_theta']))
    if 'phi' in outputs_lc:
        out_phi = np.zeros((times_array.shape[0], dist['n_phi']))
        out_phi_y = np.zeros((times_array.shape[0], dist['n_phi']))

    if 'gyro' in outputs_lc:
        out_gyro = np.zeros((times_array.shape[0], regrid[0]))
        out_gyro_y = np.zeros((times_array.shape[0], regrid[0]))

    if 'pa' in outputs_lc:
        out_pad = np.zeros((times_array.shape[0], regrid[1]))
        out_pad_y = np.zeros((times_array.shape[0], regrid[1]))

    if 'moments' in outputs_lc:
        out_density = np.zeros(times_array.shape[0])
        out_avgtemp = np.zeros(times_array.shape[0])
        out_vthermal = np.zeros(times_array.shape[0])
        out_flux = np.zeros([times_array.shape[0], 3])
        out_velocity = np.zeros([times_array.shape[0], 3])
        out_mftens = np.zeros([times_array.shape[0], 6])
        out_ptens = np.zeros([times_array.shape[0], 6])
        out_ttens = np.zeros([times_array.shape[0], 3, 3])
        out_eflux = np.zeros([times_array.shape[0], 3])
        out_t3 = np.zeros([times_array.shape[0], 3])
        out_magt3 = np.zeros([times_array.shape[0], 3])
        out_symm = np.zeros([times_array.shape[0], 3])
        out_symm_phi = np.zeros(times_array.shape[0])
        out_symm_theta = np.zeros(times_array.shape[0])
        out_symm_ang = np.zeros(times_array.shape[0])
        out_qflux = np.zeros([times_array.shape[0], 3])

    if 'fac_energy' in outputs_lc:
        out_fac_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_fac_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))

    if 'fac_moments' in outputs_lc:
        out_fac_density = np.zeros(times_array.shape[0])
        out_fac_avgtemp = np.zeros(times_array.shape[0])
        out_fac_vthermal = np.zeros(times_array.shape[0])
        out_fac_flux = np.zeros([times_array.shape[0], 3])
        out_fac_velocity = np.zeros([times_array.shape[0], 3])
        out_fac_mftens = np.zeros([times_array.shape[0], 6])
        out_fac_ptens = np.zeros([times_array.shape[0], 6])
        out_fac_ttens = np.zeros([times_array.shape[0], 3, 3])
        out_fac_eflux = np.zeros([times_array.shape[0], 3])
        out_fac_t3 = np.zeros([times_array.shape[0], 3])
        out_fac_magt3 = np.zeros([times_array.shape[0], 3])
        out_fac_symm = np.zeros([times_array.shape[0], 3])
        out_fac_symm_phi = np.zeros(times_array.shape[0])
        out_fac_symm_theta = np.zeros(times_array.shape[0])
        out_fac_symm_ang = np.zeros(times_array.shape[0])
        out_fac_qflux = np.zeros([times_array.shape[0], 3])

    out_vars = []
    last_update_time = None

    """
    ;;--------------------------------------------------------
    ;;Prepare support data
    ;;--------------------------------------------------------
    """
    # ;;create rotation matrix to B-field aligned coordinates if needed
    
    fac_outputs = ['pa','gyro','fac_energy', 'fac_moments']
    fac_requested = len(set(outputs_lc).intersection(fac_outputs)) > 0
    if fac_requested:
        """
        ;; Currently triangulation fails, so forcidly no_regrid is set for
        ;; spectum generation in FAC coordinates.
        """
        no_regrid = True

        fac_matrix = erg_pgs_make_fac(times_array, mag_name, pos_name, fac_type=fac_type)

        if fac_matrix is None:
            # problem creating the FAC matrices
            fac_requested = False

    #  ;;create the magnetic field vector array for moment calculation
    magf = np.array([0., 0., 0.])
    no_mag_for_moments = False

    if ('moments' in outputs_lc) or ('fac_moments' in outputs_lc):

        no_mag = mag_name is None
        magnm = tnames(mag_name)
        if (len(magnm) < 1) or no_mag:
            print('the magnetic field data is not given!')
            no_mag_for_moments = True
        else:
            magnm = magnm[0]

            """
            ;; Create magnetic field data with times shifted by half of spin
            ;; periods
            """

            magtmp = magnm+'_pgs_temp'
            tcopy(magnm, magtmp)
            tinterpol(magtmp, times_array, newname=magtmp)
            magf = get_data(magtmp)[1]  #  ;; [ time, 3] nT

    """
    ;;-------------------------------------------------
    ;; Loop over time to build spectrograms and/or moments
    ;;-------------------------------------------------
    """
    for index in range(time_indices.shape[0]):
        last_update_time = erg_pgs_progress_update(last_update_time=last_update_time,
             current_sample=index, total_samples=time_indices.shape[0], type_string=in_tvarname)

        #  ;; Get the data structure for this sample

        if instnm == 'lepe':

            dist = {
                'project_name': deepcopy(dist_all_time_range['project_name']),
                'spacecraft': deepcopy(dist_all_time_range['spacecraft']),
                'data_name': deepcopy(dist_all_time_range['data_name']),
                'units_name': deepcopy(dist_all_time_range['units_name']),
                'units_procedure': deepcopy(dist_all_time_range['units_procedure']),
                'species': deepcopy(dist_all_time_range['species']),
                'valid': deepcopy(dist_all_time_range['valid']),
                'charge': deepcopy(dist_all_time_range['charge']),
                'mass': deepcopy(dist_all_time_range['mass']),
                'time': deepcopy(dist_all_time_range['time'][index]),
                'end_time':  deepcopy(dist_all_time_range['end_time'][index]),
                'data':  deepcopy(dist_all_time_range['data'][:, :, :, index]),
                'bins':  deepcopy(dist_all_time_range['bins'][:, :, :, index]),
                'energy':  deepcopy(dist_all_time_range['energy'][:, :, :, index]),
                'denergy':  deepcopy(dist_all_time_range['denergy'][:, :, :, index]),
                'n_energy': deepcopy(dist_all_time_range['n_energy']),
                'n_bins': deepcopy(dist_all_time_range['n_bins']),
                'phi':  deepcopy(dist_all_time_range['phi'][:, :, :, index]),
                'dphi':  deepcopy(dist_all_time_range['dphi'][:, :, :, index]),
                'n_phi': deepcopy(dist_all_time_range['n_phi']),
                'theta':  deepcopy(dist_all_time_range['theta'][:, :, :, index]),
                'dtheta':  deepcopy(dist_all_time_range['dtheta'][:, :, :, index]),
                'n_theta': deepcopy(dist_all_time_range['n_theta'])
            }

        if instnm == 'lepi':
            dist = erg_lepi_get_dist(in_tvarname, time_indices[index], species=species, units=units_lc)

        if magf.ndim == 2:
            magvec = magf[index]
        elif magf.ndim == 1:
            magvec = magf

        if ('moments' in outputs_lc) or ('fac_moments' in outputs_lc):
            clean_data = erg_pgs_clean_data(dist, units=units_lc, magf=magvec,
                                            for_moments=True)  #;; invalid values are zero-padded. 
        else:
            clean_data = erg_pgs_clean_data(dist, units=units_lc, magf=magvec)

        if fac_requested:
            pre_limit_bins = deepcopy(clean_data['bins'])

        clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

        if ('moments' in outputs_lc) or ('fac_moments' in outputs_lc):
            clean_data_eflux = erg_convert_flux_units(clean_data, units='eflux')
            magfarr = deepcopy(magf)
            clean_data_eflux_for_moments = deepcopy(clean_data_eflux)
            clean_data_eflux_for_moments['data'] = np.where(clean_data_eflux_for_moments['bins'] == 0,
                                                            0,clean_data_eflux_for_moments['data'])
            moments = spd_pgs_moments(clean_data_eflux_for_moments)

            if 'moments' in outputs_lc:
                out_density[index] = moments['density']
                out_avgtemp[index] = moments['avgtemp']
                out_vthermal[index] = moments['vthermal']
                out_flux[index, :] = moments['flux']
                out_velocity[index, :] = moments['velocity']
                out_mftens[index, :] = moments['mftens']
                out_ptens[index, :] = moments['ptens']
                out_ttens[index, :] = moments['ttens']
                out_eflux[index, :] = moments['eflux']
                out_t3[index, :] = moments['t3']
                out_magt3[index, :] = moments['magt3']
                out_symm[index, :] = moments['symm']
                out_symm_phi[index] = moments['symm_phi']
                out_symm_theta[index] = moments['symm_theta']
                out_symm_ang[index] = moments['symm_ang']
                out_qflux[index] = moments['qflux']

        #  ;;Build theta spectrogram
        if 'theta' in outputs_lc:
            if  instnm == 'lepe':
                out_theta_y[index, :], out_theta[index, :] = erg_pgs_make_theta_spec(clean_data, no_ang_weighting=no_ang_weighting)
            elif instnm == 'lepi':
                out_theta_y[index, :], out_theta[index, :] = erg_pgs_make_theta_spec(clean_data, resolution=dist['n_theta'],no_ang_weighting=no_ang_weighting)

        #  ;;Build energy spectrogram
        if 'energy' in outputs_lc:
            out_energy_y[index, :], out_energy[index, :] = erg_pgs_make_e_spec(clean_data)

        #  ;;Build phi spectrogram
        if 'phi' in outputs_lc:
            out_phi_y[index, :], out_phi[index, :] = erg_pgs_make_phi_spec(clean_data, resolution=dist['n_phi'],no_ang_weighting=no_ang_weighting)

        #  ;;Perform transformation to FAC, (regrid data), and apply limits in new coords
        
        if fac_requested:
            
            # ;limits will be applied to energy-aligned bins
            clean_data['bins'] = deepcopy(pre_limit_bins)
            clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

            # ;perform FAC transformation and interpolate onto a new, regular grid 
            clean_data = erg_pgs_do_fac(clean_data, fac_matrix[index, :, :])

            #;nearest neighbor interpolation to regular grid in FAC
            if not no_regrid:
                if (not np.all(np.isnan(clean_data['theta']))) and (not np.all(np.isnan(clean_data['phi']))):
                    clean_data = spd_pgs_regrid(clean_data, regrid)

            clean_data['theta'] = 90.0-clean_data['theta']  #  ;pitch angle is specified in co-latitude

            # ;apply gyro & pitch angle limits(identical to phi & theta, just in new coords)
            clean_data = erg_pgs_limit_range(clean_data, theta=pitch, phi=gyro, no_ang_weighting=no_ang_weighting)

            if 'pa' in outputs_lc:
                # ;Build pitch angle spectrogram
                out_pad_y[index, :], out_pad[index, :] = erg_pgs_make_theta_spec(clean_data, colatitude=True, resolution=regrid[1], no_ang_weighting=no_ang_weighting)

            if 'gyro' in outputs_lc:
                # ;Build gyrophase spectrogram
                out_gyro_y[index, :], out_gyro[index, :] = erg_pgs_make_phi_spec(clean_data, resolution=regrid[0], no_ang_weighting=no_ang_weighting)

            if 'fac_energy' in outputs_lc:
                out_fac_energy_y[index, :], out_fac_energy[index, :] = erg_pgs_make_e_spec(clean_data)

            if 'fac_moments' in outputs_lc:
                clean_data['theta'] = 90. - clean_data['theta'] # ;convert back to latitude for moments calc
                temp_dict = {'charge': dist['charge'],
                             'magf': magvec,
                             'species': dist['species'],
                             'sc_pot': 0.,
                             'units_name': units_lc}
                temp_dict.update(clean_data)
                clean_data = deepcopy(temp_dict)
                del temp_dict
                clean_data_eflux = erg_convert_flux_units(clean_data, units='eflux')
                clean_data_eflux_for_moments = deepcopy(clean_data_eflux)
                clean_data_eflux_for_moments['data'] = np.where(clean_data_eflux_for_moments['bins'] == 0,
                                                                0,clean_data_eflux_for_moments['data'])
                fac_moments = spd_pgs_moments(clean_data_eflux_for_moments)

                out_fac_density[index] = fac_moments['density']
                out_fac_avgtemp[index] = fac_moments['avgtemp']
                out_fac_vthermal[index] = fac_moments['vthermal']
                out_fac_flux[index, :] = fac_moments['flux']
                out_fac_velocity[index, :] = fac_moments['velocity']
                out_fac_mftens[index, :] = fac_moments['mftens']
                out_fac_ptens[index, :] = fac_moments['ptens']
                out_fac_ttens[index, :] = fac_moments['ttens']
                out_fac_eflux[index, :] = fac_moments['eflux']
                out_fac_t3[index, :] = fac_moments['t3']
                out_fac_magt3[index, :] = fac_moments['magt3']
                out_fac_symm[index, :] = fac_moments['symm']
                out_fac_symm_phi[index] = fac_moments['symm_phi']
                out_fac_symm_theta[index] = fac_moments['symm_theta']
                out_fac_symm_ang[index] = fac_moments['symm_ang']
                out_fac_qflux[index] = fac_moments['qflux']

    made_et_spec = ('energy' in outputs_lc) or ('fac_energy' in outputs_lc)

    if 'energy' in outputs_lc:
        output_tplot_name = in_tvarname+'_energy' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_energy_y, z=out_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)')
        ylim(output_tplot_name,  1e+1, 3e+4) #  ;; default yrange: [10 eV, 30 keV]
        out_vars.append(output_tplot_name)
    if 'theta' in outputs_lc:
        output_tplot_name = in_tvarname+'_theta' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_theta_y, z=out_theta, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ theta (deg)')
        out_vars.append(output_tplot_name)
    if 'phi' in outputs_lc:
        output_tplot_name = in_tvarname+'_phi' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_phi_y, z=out_phi, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ phi (deg)')
        out_vars.append(output_tplot_name)

    #  ;;Pitch Angle Spectrograms
    if 'pa' in outputs_lc:
        output_tplot_name = in_tvarname+'_pa' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_pad_y, z=out_pad, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ PA (deg)')
        out_vars.append(output_tplot_name)

    if 'gyro' in outputs_lc:
        output_tplot_name = in_tvarname+'_gyro' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_gyro_y, z=out_gyro, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ gyro (deg)')
        out_vars.append(output_tplot_name)


    #  ;Moments Variables
    if 'moments' in outputs_lc:
        moments = {'density': out_density, 
              'flux': out_flux, 
              'mftens': out_mftens, 
              'velocity': out_velocity, 
              'ptens': out_ptens,
              'ttens': out_ttens,
              'vthermal': out_vthermal,
              'avgtemp': out_avgtemp,
              'eflux': out_eflux,
              't3': out_t3,
              'magt3': out_magt3,
              'symm': out_symm,
              'symm_phi': out_symm_phi,
              'symm_theta': out_symm_theta,
              'symm_ang': out_symm_ang,
              'qflux': out_qflux,
           }
        moments_vars = erg_pgs_moments_tplot(moments, x=times_array, prefix=in_tvarname, suffix=suffix)
        out_vars.extend(moments_vars)

    if 'fac_energy' in outputs_lc:

        output_tplot_name = in_tvarname+'_energy_mag' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_fac_energy_y, z=out_fac_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)')
        ylim(output_tplot_name, 1e+1, 3e+4)  # ;; default yrange: [10 eV, 30 keV]
        out_vars.append(output_tplot_name)

    #  ;FAC Moments Variables
    if 'fac_moments' in outputs_lc:
        fac_moments = {'density': out_fac_density, 
              'flux': out_fac_flux, 
              'mftens': out_fac_mftens, 
              'velocity': out_fac_velocity, 
              'ptens': out_fac_ptens,
              'ttens': out_fac_ttens,
              'vthermal': out_fac_vthermal,
              'avgtemp': out_fac_avgtemp,
              'eflux': out_fac_eflux,
              't3': out_fac_t3,
              'magt3': out_fac_magt3,
              'symm': out_fac_symm,
              'symm_phi': out_fac_symm_phi,
              'symm_theta': out_fac_symm_theta,
              'symm_ang': out_fac_symm_ang,
              'qflux': out_fac_qflux,
                       }
        fac_mom_suffix = '_mag' + suffix
        fac_moments_vars = erg_pgs_moments_tplot(fac_moments, x=times_array, prefix=in_tvarname, suffix=fac_mom_suffix)
        out_vars.extend(fac_moments_vars)

    
    #  ;;Sort a data array by energy for (fac-)energy spectra
    if ('erg_lepe_' in in_tvarname)  and (made_et_spec):
        if 'energy' in outputs_lc:
            t_plot_name = in_tvarname+'_energy' + suffix
            get_data_energy = get_data(t_plot_name)
            energy_meta_data = get_data(t_plot_name, metadata=True)
        elif 'fac_energy' in outputs_lc:
            t_plot_name = in_tvarname+'_energy_mag' + suffix
            get_data_energy = get_data(t_plot_name)
            energy_meta_data = get_data(t_plot_name, metadata=True)
        
        if get_data_energy is not None:
            
            arange_time_indices = np.arange(get_data_energy[0].size)
            time_indices_repeat = np.repeat(np.array([arange_time_indices]).T, get_data_energy[1].shape[1], axis=1)
            time_indices_repeat_reshape =  time_indices_repeat.reshape((time_indices_repeat.size, 1))
            
            if get_data_energy[2].ndim == 1:
                arg_sort_axis_1 = np.repeat(np.argsort([get_data_energy[2]], axis=1), get_data_energy[0].size, axis=0)
            elif get_data_energy[2].ndim == 2:
                arg_sort_axis_1=np.argsort(get_data_energy[2], axis=1)

            arg_sort_axis_1_reshape = arg_sort_axis_1.reshape((arg_sort_axis_1.size, 1))

            indices_array =np.concatenate([time_indices_repeat_reshape, arg_sort_axis_1_reshape], axis=1)
            indices_list_0 = indices_array[:,0].tolist()
            indices_list_1 = indices_array[:,1].tolist()
            y_new_1d = get_data_energy[1][indices_list_0 , indices_list_1 ]
            y_new_2d = y_new_1d.reshape(get_data_energy[1].shape)
            v_new_1d = get_data_energy[2][indices_list_0 , indices_list_1]
            v_new_2d = v_new_1d.reshape(get_data_energy[2].shape)

            store_data(t_plot_name, data={'x':get_data_energy[0],
                                          'y':y_new_2d,
                                          'v':v_new_2d},
                        attr_dict=energy_meta_data)

    return out_vars



"""
Removes NaNs and other flags.

Notes
----
Allowed wildcards are ? for a single character, * from multiple characters.
Similar to tdeflag.pro in IDL SPEDAS, but now a wrapper for deflag.py

"""
import logging
import pyspedas
from pyspedas.tplot_tools import tnames, deflag

def tdeflag(names,
            flag=None,
            method='remove_nan',
            newname=None,
            suffix=None,
            overwrite=None,
            fillval=None
):
    """
    Replaces FLAGs in arrays with interpolated or other values.

    Parameters
    ----------
    names: str/list of str
        List of tplot variables names to deflag (wildcards accepted)
    flag: float, int, or list
        Value or values to be treated as flags
    method: str, optional
        Method to apply.
        If None,then flagged values are replaced with NaN
        Other options 'repeat' (repeat last good value), 
        'linear' (interpolate linearly over gap).
        'replace' replaces flagged value with a fill value, which can be set using the 
                  keyword 'fillval' (default is to use NaN)
        Option 'remove_nan' removes time intervals with NaN values
        Default: 'remove_nan'
    newname: str/list of str, optional
        List of new names for tplot variables.
        If '', then tplot variables are replaced.
        Default: None. If not specified then a suffix is applied.
    suffix: str, optional
        A suffix to apply.
        Default: '-deflag'.
    overwrite: bool, optional
        Replace the existing tplot name.
        Default: None
    fillval: int, float (optional)
        Value to use as replacement if method='replace'

    Returns
    -------
    list of str
        List of tplot variables created or altered by tdeflag

    Examples
    --------

    >>> import pyspedas
    >>> pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,4],[4,90],[5,5],[6,6],[7,7]]})
    >>> # Remove any instances of [100,90,7,2,57] from 'd', store in 'e'.
    >>> pyspedas.deflag('d',[100,90,7,2,57],newname='e')
    >>> d_deflagged = pyspedas.get_data('e')
    >>> print(d_deflagged)
    """

    old_names = tnames(names)

    if len(old_names) < 1:
        logging.error('tdeflag: No valid tplot names were provided.')
        return None

    if suffix is None:
        suffix = '-deflag'

    if overwrite is not None:
        n_names = old_names
    elif newname is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]
        logging.info('input newname has incorrect number of elements')

    for i in range(len(old_names)):
        deflag(old_names[i], flag, newname=n_names[i], method=method, fillval=fillval)
        logging.info('tdeflag was applied to: ' + n_names[i])

    return n_names



from pyspedas.tplot_tools import time_string, time_double
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.hapi_tools.hapi import hapi
from .config import CONFIG

def load(trange=['2017-03-27/06:00', '2017-03-27/08:00'], 
         probe='a',
         instrument='mag',
         datatype='', 
         level='l1b', 
         prefix='',
         suffix='',
         varnames=[],  # should be sent to hapi as parameters?
         time_clip=False,
         force_download=False):
    """
    Loads data from the Swarm mission.

    This function is not meant to be called directly; instead, see the wrappers such as:
        - pyspedas.projects.swarm.mag.

    Parameters
    ----------
    trange : list of str, default=['2017-03-27/06:00', '2017-03-27/08:00']
        Time range of interest [starttime, endtime] with the format 'YYYY-MM-DD' or 'YYYY-MM-DD/hh:mm:ss'.
    probe : str or list of str, default='a'
        Swarm spacecraft ID(s) ('a', 'b', and/or 'c').
    instrument : str, default='mag'
        The instrument from which to load data. For magnetometer data, use 'mag'.
    datatype : str, optional
        Specific data type to load, if applicable. For magnetometer data, options include 'hr' (high resolution) and 'lr' (low resolution).
    level : str, default='l1b'
        Data level; for magnetometer data. Options: 'l1b'.
    prefix : str, optional
        The tplot variable names will be given this prefix. By default, no prefix is added.
    suffix : str, optional
        The tplot variable names will be given this suffix. By default, no suffix is added.
    varnames : list of str, optional
        List of variable names to load as parameters. If not specified, all data variables are loaded.
    time_clip : bool, default=False
        Time clip the variables to exactly the range specified in the trange keyword.
    force_download : bool, default=False
        Swarm data is requested via HAPI. This parameter always ignored and reserved for compatibility.

    Returns
    -------
    out_vars : list of str
        List of tplot variable names created by the data loading process.

    Examples
    --------
    This function is not intended to be called directly.
    """


    server = CONFIG['remote_data_dir'] + 'hapi/'

    if not isinstance(probe, list):
        probe = [probe]

    tr = time_string(time_double(trange), fmt='%Y-%m-%dT%H:%M:%S')

    out_vars = []

    if prefix:
        prefix2 = prefix
    else:
        prefix2 = ''

    for this_probe in probe:
        prefix = prefix2 + 'swarm' + this_probe.lower() + '_'
        if instrument == 'mag':
            dataset = 'SW_OPER_MAG' + this_probe.upper() + '_' + datatype.upper() + '_' + level[1:].upper()
            tvars = hapi(trange=tr, 
                         server=server, 
                         dataset=dataset, 
                         parameters=varnames,
                         suffix=suffix,
                         prefix=prefix)

            if tvars is not None and len(tvars) != 0:
                out_vars.extend(tvars)

    if time_clip:
        for new_var in out_vars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return out_vars



from .gmag_isee_fluxgate import gmag_isee_fluxgate

from typing import List, Union, Optional, Dict, Any

def gmag_stel_fluxgate(
    trange: List[str] = ['2020-08-01', '2020-08-02'],
    suffix: str = '',
    site: Union[str, List[str]] = 'all',
    datatype: Union[str, List[str]] = 'all',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> Union[Dict, None, List[Union[str, Any]]]:
    """
    Load data from STEL Fluxgate Magnetometers (wrapper for gmag_isee_fluxgate)

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-08-01', '2020-08-02']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load.
            Valid values: 'msr', 'rik', 'kag', 'ktb', 'lcl', 'mdm', 'tew', 'all'
            Default: ['all']

    datatype: str or list of str
            The data types to load. Valid values: '64hz', '1sec', '1min', '1h', 'all'
            Default: 'all'

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> fluxgate_vars=pyspedas.projects.erg.gmag_stel_fluxgate(trange=['2020-08-01','2020-08-02'], site='all')
    >>> tplot('isee_fluxgate_mag_ktb_1min_hdz')


    """

    return gmag_isee_fluxgate(
        trange=trange,
        suffix=suffix,
        site=site,
        datatype=datatype,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        no_update=no_update,
        uname=uname,
        passwd=passwd,
        time_clip=time_clip,
        ror=ror,
        force_download=force_download,
    )



from pyspedas import clip, options, ylim, zlim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def pwe_hfa(
    trange: List[str] = ['2017-04-01', '2017-04-02'],
    datatype: str = 'spec',
    mode: str = 'low',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-04-01', '2017-04-02']

        datatype: str
            Data type; Valid 'l2' options: 'spec'
            Default: 'spec'

        level: str
            Data level; Valid options: 'l2', 'l3'
            Default: 'l2'

        mode: str
            Mode of 'l2' data to load. Valid options: ''high', 'low', 'monit'
            Default: 'low'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.  Default: None (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road
            Default: True

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> pwe_hfa_vars = pyspedas.projects.erg.pwe_hfa(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_pwe_hfa_l2_low_spectra_eu')

    """

    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24

    if level == 'l2':
        prefix = 'erg_pwe_hfa_'+level+'_' + mode + '_'

    if level == 'l2':
        pathformat = 'satellite/erg/pwe/hfa/'+level+'/'+datatype+'/'+mode + \
            '/%Y/%m/erg_pwe_hfa_'+level+'_'+datatype+'_'+mode+'_%Y%m%d_v??_??.cdf'
    elif level == 'l3':
        prefix = 'erg_pwe_hfa_'+level+'_1min_'
        pathformat = 'satellite/erg/pwe/hfa/'+level + \
            '/%Y/%m/erg_pwe_hfa_'+level+'_1min_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE HFA')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ", gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/HFA: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Hfa')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if (level == 'l2') and (mode == 'low') and (not notplot):

        # set spectrogram plot option
        options(prefix + 'spectra_eu' + suffix, 'Spec', 1)
        options(prefix + 'spectra_ev' + suffix, 'Spec', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'Spec', 1)
        options(prefix + 'spectra_esum' + suffix, 'Spec', 1)
        options(prefix + 'spectra_er' + suffix, 'Spec', 1)
        options(prefix + 'spectra_el' + suffix, 'Spec', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'Spec', 1)
        options(prefix + 'spectra_e_ar' + suffix, 'Spec', 1)

        if prefix + 'spectra_er' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_er' + suffix, 0., 5000.)
        if prefix + 'spectra_el' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_el' + suffix, 0., 5000.)

        if prefix + 'spectra_eu' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_eu' + suffix, 0., 5000.)
            # set ylim
            ylim(prefix + 'spectra_eu' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_eu' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_ev' + suffix in loaded_data:
            # remove minus values in y array
            clip(prefix + 'spectra_ev' + suffix, 0., 5000.)
            # set ylim
            ylim(prefix + 'spectra_ev' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_ev' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_bgamma' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_bgamma' + suffix, 2.0, 200.0)
            # set zlim
            zlim(prefix + 'spectra_bgamma' + suffix, 1e-4, 1e+2)

        if prefix + 'spectra_esum' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_esum' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_esum' + suffix,  1e-10, 1e-3)

        if prefix + 'spectra_e_ar' + suffix in loaded_data:
            # set ylim
            ylim(prefix + 'spectra_e_ar' + suffix,  2.0, 10000.0)
            # set zlim
            zlim(prefix + 'spectra_e_ar' + suffix, -1, 1)

        # set y axis to logscale
        options(prefix + 'spectra_eu' + suffix, 'ylog', 1)
        options(prefix + 'spectra_ev' + suffix, 'ylog', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'ylog', 1)
        options(prefix + 'spectra_esum' + suffix, 'ylog', 1)
        options(prefix + 'spectra_er' + suffix, 'ylog', 1)
        options(prefix + 'spectra_el' + suffix, 'ylog', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'ylog', 1)
        options(prefix + 'spectra_e_ar' + suffix, 'ylog', 1)

        # set z axis to logscale
        options(prefix + 'spectra_eu' + suffix, 'zlog', 1)
        options(prefix + 'spectra_ev' + suffix, 'zlog', 1)
        options(prefix + 'spectra_bgamma' + suffix, 'zlog', 1)
        options(prefix + 'spectra_esum' + suffix, 'zlog', 1)
        options(prefix + 'spectra_er' + suffix, 'zlog', 1)
        options(prefix + 'spectra_el' + suffix, 'zlog', 1)
        options(prefix + 'spectra_e_mix' + suffix, 'zlog', 1)

        # set ytitle
        options(prefix + 'spectra_eu' + suffix, 'ytitle', 'ERG PWE/HFA (EU)')
        options(prefix + 'spectra_ev' + suffix, 'ytitle', 'ERG PWE/HFA (EV)')
        options(prefix + 'spectra_esum' + suffix,
                'ytitle', 'ERG PWE/HFA (ESUM)')
        options(prefix + 'spectra_e_ar' + suffix,
                'ytitle', 'ERG PWE/HFA (E_AR)')
        options(prefix + 'spectra_bgamma' + suffix,
                'ytitle', 'ERG PWE/HFA (BGAMMA)')

        # set ysubtitle
        options(prefix + 'spectra_eu' + suffix, 'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_ev' + suffix, 'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_esum' + suffix,
                'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_e_ar' + suffix,
                'ysubtitle', 'frequency [Hz]')
        options(prefix + 'spectra_bgamma' + suffix,
                'ysubtitle', 'frequency [Hz]')

        # set ztitle
        options(prefix + 'spectra_eu' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_ev' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_esum' + suffix, 'ztitle', 'mV^2/m^2/Hz')
        options(prefix + 'spectra_e_ar' + suffix, 'ztitle', 'LH:-1/RH:+1')
        options(prefix + 'spectra_bgamma' + suffix, 'ztitle', 'pT^2/Hz')

        # change colormap option
        options(prefix + 'spectra_eu' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_ev' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_bgamma' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_esum' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_er' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_el' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_e_mix' + suffix, 'Colormap', 'jet')
        options(prefix + 'spectra_e_ar' + suffix, 'Colormap', 'jet')

    elif level == 'l3':

        # set ytitle
        options(prefix + 'Fuhr' + suffix, 'ytitle', 'UHR frequency [Mhz]')
        options(prefix + 'ne_mgf' + suffix,
                'ytitle', 'eletctorn density [/cc]')

        # set y axis to logscale
        options(prefix + 'Fuhr' + suffix, 'ylog', 1)
        options(prefix + 'ne_mgf' + suffix, 'ylog', 1)

    return loaded_data
